/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "list.h"
#include "mapfile.h"
#include "database.h"

#define VERSION 3

#define GLOBALINDEX 4
#define STATICINDEX 12
#define HASHINDEX 28
#define DYNAMICINDEX 36
#define HEADERLENGTH 44

#define readInt(buf, x) (*(x) = (buf)[0] + ((buf)[1] << 8) + ((buf)[2] << 16) + ((buf)[3] << 24))
#define badVal(offset) (((offset)<0) || ((offset)>=gdaLength))
#define badOffsetLength(offset, length) ((badVal(offset)) || (badVal(offset + length)) || (length < 0))

int counter = 0; /* nb of calls to getDepInfos */

static unsigned char *gdaBuffer = 0;
static int gdaLength = 0;
static int staticOffset = 0;
static int staticLength = 0;
static int hashOffset = 0;
static int hashSize = 0;

void map(char *file) {
  struct map map[1];
  int version, hashLength;
  
  mapFile(file, map);
  gdaLength = map->length;
  gdaBuffer = map->buffer;
  
  if(gdaLength < HEADERLENGTH) fatal("incorrect header");

  readInt(gdaBuffer, &version);
  if(version != VERSION) fatal("wrong version");
  readInt(gdaBuffer+STATICINDEX, &staticOffset);
  readInt(gdaBuffer+STATICINDEX+4, &staticLength);
  if(badOffsetLength(staticOffset, staticLength)) fatal("the static segment is corrupted");
  readInt(gdaBuffer+HASHINDEX, &hashOffset);
  readInt(gdaBuffer+HASHINDEX+4, &hashLength);
  if(badOffsetLength(hashOffset, hashLength)) fatal("the target segment is corrupted");
  hashSize = hashLength / 4;
}

static int skipString(int offset, int limit) {
  for(; offset<limit && gdaBuffer[offset]; ++offset);
  if(++offset > limit) fatal("a string is corrupted");
  return offset;
}

void getVarInfos(int offset, struct varInfos *infos) {
  int i, str, limit;

  if(badOffsetLength(offset, 16)) goto varError;
  readInt(gdaBuffer+offset, &i);
  readInt(gdaBuffer+offset+4, &limit);
  if(badOffsetLength(i, limit-i)) goto varError;  
  infos->name = (char*)gdaBuffer+i;
  i = skipString(i, limit);
  if(i+12 > limit) goto varError;
  readInt(gdaBuffer+i, &str);
  if(badVal(str)) goto varError;
  infos->loc.file = (char*)gdaBuffer + str;
  readInt(gdaBuffer+i+4, &infos->loc.startLine);
  readInt(gdaBuffer+i+8, &infos->loc.endLine);  
  i += 12;
  infos->ctype = (char*)gdaBuffer+i;
  i = skipString(i, limit);
  if(i+4 > limit) goto varError;
  readInt(gdaBuffer+i, &infos->function);  

  return;

 varError:
  fatal("a variable is corrupted");
}

void getVarOffsets(int offset, struct varOffsets *offsets) {
  int nb;

  if(badOffsetLength(offset, 16)) goto varError;
  readInt(gdaBuffer+offset+4, &offsets->forward);
  readInt(gdaBuffer+offset+8, &offsets->backward);
  readInt(gdaBuffer+offset+12, &offsets->end);
  offsets->prototype = 0;
  if(badOffsetLength(offset, offsets->end-offset)) goto varError;

  if(offsets->forward < offsets->backward && isPrototype(gdaBuffer[offsets->forward])) {
    readInt(gdaBuffer+offsets->forward+5, &nb);
    offsets->prototype = offsets->forward;
    offsets->forward += 9 + 4 * nb;
  }
  
  return;

 varError:
  fatal("a variable is corrupted");
}

int getDepInfos(int offset, struct depInfos *infos) {
  int str;

  counter++;
  if(badOffsetLength(offset, 17)) goto depError;
  infos->dtype = gdaBuffer[offset];
  readInt(gdaBuffer+offset+1, &infos->var);
  readInt(gdaBuffer+offset+5, &str);
  if(badVal(str)) goto depError;
  infos->loc.file = (char*)gdaBuffer + str;
  readInt(gdaBuffer+offset+9, &infos->loc.startLine);
  readInt(gdaBuffer+offset+13, &infos->loc.endLine);  
  offset += 17;
  infos->lstr = (char*)gdaBuffer + offset;
  offset = skipString(offset, gdaLength);
  infos->rstr = (char*)gdaBuffer + offset;
  offset = skipString(offset, gdaLength);

  return offset;
  
 depError:
  fatal("a dependency is corrupted");
  return 0;
}

unsigned char getProtoInfos(int offset, struct list *offsets) {
  int i, nb, val;

  if(!offsets) return gdaBuffer[offset];
  readInt(gdaBuffer+offset+1, &val);
  readInt(gdaBuffer+offset+5, &nb);
  *((int *)&offsets->elem) = val;
  for(i=0; i<nb; i++) {
    offsets = offsets->next = newCons();
    readInt(gdaBuffer+offset+9+4*i, &val);
    *((int *)&offsets->elem) = val;
  }
  offsets->next = 0;
  return gdaBuffer[offset];
}

int getStaticInfos(int offset, struct staticInfos *infos) {
  int str;

  if(!offset) offset = staticOffset;
  if(offset >= staticOffset + staticLength) return 0;
  if(badVal(offset)) goto staticError;
  infos->dtype = gdaBuffer[offset];
  switch(infos->dtype) {
  case '&':
    if(badOffsetLength(offset, 21)) goto staticError;
    readInt(gdaBuffer+offset+1, &infos->var);
    readInt(gdaBuffer+offset+5, &infos->lval);
    readInt(gdaBuffer+offset+9, &str);
    if(badVal(str)) goto staticError;
    infos->loc.file = (char*)gdaBuffer + str;
    readInt(gdaBuffer+offset+13, &infos->loc.startLine);
    readInt(gdaBuffer+offset+17, &infos->loc.endLine);
    offset += 21;
    break;
    
  default:
    goto staticError;
  }
  return offset;
  
 staticError:
  fatal("a static dependency is corrupted");
  return 0;
}

int findTarget(char *name, struct varInfos *infos) {
  static int hash, around;
  int i, target;

  if(name) {
    around = hash = 0;
    for(i=0; name[i]; i++) hash = ((hash * 67) + name[i]) % hashSize;
  }
  readInt(gdaBuffer+hashOffset+hash*4, &target);
  if(!target) return 0;
  getVarInfos(target, infos);
  hash += 17;
  if(hash >= hashSize) {
    if(around++ >= 17) fatal("hash table wrap around??");
    hash %= hashSize;
  }
  return target;
}
