/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "list.h"
#include "storage.h"
#include "set.h"
#include "database.h"
#include "aliasing.h"
#include "analyse.h"
#include "targets.h"
#include "print.h"

static int direction = FORWARD;
static int outputLength = 20;
static struct entry pattern[1];

static int input(char *buf, int size, FILE *stream) {
  int i, state = 1;
  
  for(;;) {
    if(!fgets(buf, size, stream)) return -1;
    for(i=0; buf[i]; i++);
    if(buf[--i] == '\n') {
      buf[i] = 0;
      return state; /* 0: line too long, 1: ok, -1: EOF */
    } else {
      state = 0;
    }
  }
}

static void quit(void) {
  fprintf(stderr, "Bye\n");
  exit(0);
}

static void toplevel(char *buf, int size) {
  for(;;) {
    fprintf(stderr, "? ");
    switch(input(buf, size, stdin)) {
    case -1: fprintf(stderr, "\n"); quit();
    case 0: fprintf(stderr, "Error: line too long\n"); break;
    case 1: return;
    }
  }
}

static void error(void) {
  fprintf(stderr, "Error: incorrect command\n");
}

static void show(void) {
  char *flag;
  
  while(flag = strtok(0, " ")) {
    if(!strcmp(flag, "locations")) {
      showLocations = 1;
    } else if(!strcmp(flag, "anonymous")) {
      showAnonymousVars = 1;
    } else if(!strcmp(flag, "types")) {
      showTypes = 1;
    } else if(!strcmp(flag, "strings")) {
      showStrings = 1;
    } else {
      error();
      return;
    }
  }
}

static void hide(void) {
  char *flag;
  
  while(flag = strtok(0, " ")) {
    if(!strcmp(flag, "locations")) {
      showLocations = 0;
    } else if(!strcmp(flag, "anonymous")) {
      showAnonymousVars = 0;
    } else if(!strcmp(flag, "types")) {
      showTypes = 0;
    } else if(!strcmp(flag, "strings")) {
      showStrings = 0;
    } else {
      error();
      return;
    }
  }
}

static void listTargets(void) {
  int i;
  
  fprintf(stderr, "%d target variable(s)", targetCount());
  if(targetCount()) fprintf(stderr, ": ");
  for(i=0; i<targetCount(); i++) fprintf(stderr, "%s ", targetOfIndex(i)->name);
  fprintf(stderr, "\n");
}

static void listNonTargets(void) {
  int i;
  
  fprintf(stderr, "%d non target variable(s)", nonTargetCount());
  if(nonTargetCount()) fprintf(stderr, ": ");
  for(i=0; i<nonTargetCount(); i++) fprintf(stderr, "%s ", nonTargetOfIndex(i)->name);
  fprintf(stderr, "\n");
}

static void go(void) {
  listTargets();
  listNonTargets();
  fprintf(stderr, "Analysing");
  switch(direction) {
  case FORWARD: fprintf(stderr, " (forward)"); break;
  case BACKWARD: fprintf(stderr, " (backward)"); break;
  case UNIFICATION: fprintf(stderr, " (unification)");
  }
  restoreState();
  analyse(direction);
  fprintf(stderr, "\n%d dependent variable(s)\n", depVarCount());
}

static void forward(void) {
  direction = FORWARD;
}

static void backward(void) {
  direction = BACKWARD;
}

static void unification(void) {
  direction = UNIFICATION;
}

static void reset(void) {
  resetTargets();
  listTargets();
  listNonTargets();
}

static void limit(void) {
  char *str = strtok(0, " ");
  int l;
  
  if(str && sscanf(str, "%d", &l)) outputLength = l; else error();
}

static void load(int mode) {
  char *name = strtok(0, " ");

  if(name) {
    importTargets(name, mode);
  } else {
    error();
  }
}

static void seed(void) {
  char *line = strtok(0, "");

  if(line) parseTargets(line, strlen(line) + 1, TARGETS);
  listTargets();
}

static void skip(void) {
  char *line = strtok(0, "");
  
  if(line) parseTargets(line, strlen(line) + 1, NONTARGETS);
  listNonTargets();
}

static void usage(void) { 
  fprintf(stderr, "usage: gda - \"file.gda\" for interactive mode\n");
  fprintf(stderr, "   or: gda \"file.gda\" [\"targets\" [\"nontargets\"]] in batch mode\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "flags in batch mode:\n");
  fprintf(stderr, " -q:  quiet mode\n");
  fprintf(stderr, " -qq: quiet mode\n");
  fprintf(stderr, " -nr: no results\n");
  fprintf(stderr, "\n");
  exit(1);
}

static void batch(int argc, char *argv[]) {
  int nr = 0, quiet = 0, index = 0, i;

  for(i=1; i<argc; i++) {
    if(!strcmp(argv[i], "-q")) { quiet = 1; continue; }
    if(!strcmp(argv[i], "-qq")) { quiet = 1; continue; }
    if(!strcmp(argv[i], "-nr")) { nr = 1; continue; }
    break;
  }
  if(i == argc || i < argc - 3) usage();
  initialize();
  map(argv[i++]);
  aliasing();
  if(debug) {
    fprintf(stderr, "%d deps\n", depCount());
    fprintf(stderr, "%d vars\n", varCount());
    fprintf(stderr, "%d deps II\n", counter);
  }
  saveState();

  if(!quiet) fprintf(stderr, "%d lval(s)\n", lvalCount());
  
  if(i == argc) {
    if(!nr) printVars(LVALS, &index, 0, pattern);
    exit(0);
  }

  initializeTargets();
  importTargets(argv[i++], TARGETS);
  if(!quiet) listTargets();

  if(i < argc) {
    importTargets(argv[i++], NONTARGETS);
    if(!quiet) listNonTargets();
  }
  
  analyse(direction);
  if(!quiet) fprintf(stderr, "%d dependent variable(s)\n", depVarCount());

  if(!nr) printVars(CHAINS, &index, 0, pattern);
  exit(0);
}

static void filter(void) {
  char *line = strtok(0, "");

  if(line && !parsePattern(line, pattern)) {
    fprintf(stderr, "Error: incorrect pattern\n");
    return;
  }
  fprintf(stderr, "Pattern: ");
  fprintPattern(stderr, pattern);
  fprintf(stderr, "\n");
}

static void printMany(int mode) {
  static int index = 0;
  static int last = NOVARS;
  
  if(mode != NOVARS) {
    index = 0;
    last = mode;
  }
  printVars(last, &index, outputLength, pattern);
}

static void next(void) {
  int index = 0;
  struct entry *entry;

  entry = printVars(DEPS, &index, 2, pattern);
  if(entry) {
    copyPattern(entry, pattern);
    fprintf(stderr, "Next: ");
    fprintPattern(stderr, pattern);
    fprintf(stderr, "\n");
  }
}

static void help(void) {
  fprintf(stderr, "Commands:\n");
  fprintf(stderr, "help\n");
  fprintf(stderr, "quit\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "show [strings] [types] [locations] [anonymous]\n");
  fprintf(stderr, "hide [strings] [types] [locations] [anonymous]\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "seed [\"pattern\"] ...\n");
  fprintf(stderr, "skip [\"pattern\"] ...\n");
  fprintf(stderr, "load \"file\"\n");
  fprintf(stderr, "loadn \"file\"\n");
  fprintf(stderr, "reset\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "limit \"number\"\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "forward\n");
  fprintf(stderr, "backward\n");
  fprintf(stderr, "unification\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "go\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "filter [\"pattern\"]\n");
  fprintf(stderr, "next\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "allvars\n");
  fprintf(stderr, "depvars\n");
  fprintf(stderr, "targets\n");
  fprintf(stderr, "nontargets\n");
  fprintf(stderr, "lvals\n");
  fprintf(stderr, "chains\n");
  fprintf(stderr, "deps\n");
  fprintf(stderr, "revdeps\n");
  fprintf(stderr, "\n");
}

#define MAXLINE 100

int main(int argc, char *argv[]) {
  char line[MAXLINE];
  char *command;
  
  initializeStream();
  if(argc < 2) usage();
  
  if(strcmp(argv[1], "-")) batch(argc, argv);
  
  if(argc != 3) usage();

  map(argv[2]);
  fprintf(stderr, "Aliasing analysis");
  initialize();
  aliasing();
  saveState();
  fprintf(stderr, "\n%d lval(s)\n", lvalCount());
  
  initializeTargets();

  for(;;) {
    toplevel(line, MAXLINE);
    command = strtok(line, " ");
    
    if(!command) {
      printMany(NOVARS);
    }
    else if(!strcmp(command, "help")) help();
    else if(!strcmp(command, "quit")) quit();

    else if(!strcmp(command, "show")) show();
    else if(!strcmp(command, "hide")) hide();

    else if(!strcmp(command, "seed")) seed(); /* fix */
    else if(!strcmp(command, "skip")) skip(); /* fix */
    else if(!strcmp(command, "load")) load(TARGETS);
    else if(!strcmp(command, "loadn")) load(NONTARGETS);
    else if(!strcmp(command, "reset")) reset();

    else if(!strcmp(command, "limit")) limit();

    else if(!strcmp(command, "forward")) forward();
    else if(!strcmp(command, "backward")) backward();
    else if(!strcmp(command, "unification")) unification();

    else if(!strcmp(command, "go")) go();

    else if(!strcmp(command, "filter")) filter();
    else if(!strcmp(command, "next")) next();
    
    else if(!strcmp(command, "allvars")) printMany(ALLVARS);
    else if(!strcmp(command, "depvars")) printMany(DEPVARS);
    else if(!strcmp(command, "targets")) printMany(TARGETVARS);
    else if(!strcmp(command, "nontargets")) printMany(NONTARGETVARS);
    else if(!strcmp(command, "lvals")) printMany(LVALS);
    else if(!strcmp(command, "chains")) printMany(CHAINS);
    else if(!strcmp(command, "deps")) printMany(DEPS);
    else if(!strcmp(command, "revdeps")) printMany(REVDEPS);
    else if(!strcmp(command, "lvalsdeps")) printMany(LVALSDEPS);
    else error();
  }
}
