/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "mapfile.h"
#include "list.h"
#include "registry.h"
#include "readGda.h"

#define VERSION 3
#define PROTOTYPE 0x80

unsigned char *buf = 0;
int pos = 0;
int length = 0;

static char *nextString(void) {
  int i = pos;

  while(buf[pos++] && pos < length);
  if(pos > length) fatal("string corrupted");
  return (char *)buf+i;
}

#define CHECK(pos, len) ((pos<0||pos+len>length)?fatal("file corrupted"):0)
#define NEXTINT() (CHECK(pos, 4), pos+=4, (buf[pos-4]) + (buf[pos-3] << 8) + (buf[pos-2] << 16) + (buf[pos-1] << 24))
#define NEXTCHAR() (CHECK(pos, 1), buf[pos++])
#define NEXTSTRING() (CHECK(pos, 1), nextString())

static struct dependency *newDep;
#define newDep(dep) (newDep = newDependency(), newDep->next = (dep), (dep) = newDep)

static void globalSection(int len) {
  char *name;
  
  while(pos < len) {
    name = insertStr(NEXTSTRING());
    findGlobal(NEXTINT(), name);
  }
}

static void stringsSection(int len) {
  int currentPos;

  while(pos < len) {
    currentPos = pos;
    findLocalStr(currentPos)->name = insertStr(NEXTSTRING());
  }
}

static void staticSection(int len) {
  struct dependency *dep;
  unsigned char dtype;
  int offset;

  while(pos < len) {
    dtype = NEXTCHAR();
    offset = NEXTINT();
    dep = newDep(findLocalVar(offset)->staticDeps);
    dep->dtype = dtype;
    dep->variable = findLocalVar(NEXTINT());
    dep->locString = findLocalStr(NEXTINT())->name;
    if(!dep->locString) fatal("strings corrupted");
    dep->startLine = NEXTINT();
    dep->endLine = NEXTINT();
  }
}

static void dynamicSection(int len) {
  int commonOffset, depOffset, revOffset, endOffset;
  struct variable *var;
  struct dependency *dep;
  int function, ret, i, nb;
  unsigned char dtype;

  while (pos < len) {
    var = findLocalVar(pos);
    commonOffset = NEXTINT();
    depOffset = NEXTINT();
    revOffset = NEXTINT();
    endOffset = NEXTINT();
    
    pos = commonOffset;
    if(!var->name) {
      var->name = STRDUP(NEXTSTRING());
    } else {
      /* global variable, we already have the name */
      NEXTSTRING();
    }
    var->locString = findLocalStr(NEXTINT())->name;
    if(!var->locString) fatal("strings corrupted");
    var->startLine = NEXTINT();
    var->endLine = NEXTINT();
    if(!var->type) {
      var->type = STRDUP(NEXTSTRING());
    } else {
      /* could check that old and new types are identical */
      NEXTSTRING();
    }
    function = NEXTINT();
    if(function) {
      var->function = findLocalVar(function);
    }

    pos = depOffset;

    while(pos < revOffset) {
      dtype = NEXTCHAR();
      if(dtype & PROTOTYPE) {
	ret = NEXTINT();
	nb = NEXTINT();
	if(!var->prototype) {
	  /* ignore functions with multiple prototypes */
	  var->prototype = (struct prototype *)MALLOC(nb * sizeof(struct variable *) + sizeof(struct prototype));
	  var->prototype->dtype = dtype;
	  if(ret) {
	    var->prototype->ret = findLocalVar(ret);
	  } else {
	    var->prototype->ret = 0;
	  }
	  var->prototype->nb = nb;
	  for (i=0; i<nb; i++) {
	    *(((struct variable **)(var->prototype+1))+i) = findLocalVar(NEXTINT());
	  }
	} else {
	  for (i=0; i<nb; i++) NEXTINT();
	}
      } else {
	dep = newDep(var->deps);
	dep->dtype = dtype;
	dep->variable = findLocalVar(NEXTINT());
	dep->locString = findLocalStr(NEXTINT())->name;
	if(!dep->locString) fatal("strings corrupted");
	dep->startLine = NEXTINT();
	dep->endLine = NEXTINT();
	dep->stringL = STRDUP(NEXTSTRING());
	dep->stringR = STRDUP(NEXTSTRING());
      }
    }
    while(pos < endOffset) {
      dep = newDep(var->revDeps);
      dep->dtype = NEXTCHAR();
      dep->variable = findLocalVar(NEXTINT());
      dep->locString = findLocalStr(NEXTINT())->name;
      if(!dep->locString) fatal("strings corrupted");
      dep->startLine = NEXTINT();
      dep->endLine = NEXTINT();
      dep->stringL = STRDUP(NEXTSTRING());
      dep->stringR = STRDUP(NEXTSTRING());
    }
  }
}

int readGda(char *file) {
  static int warning = 0;
  struct map map[1];
  int version;
  int globalOffset, globalLength;
  int staticOffset, staticLength;
  int stringsOffset, stringsLength; 
  int dynamicOffset, dynamicLength;

  mapFile(file, map);
  fprintf(stderr, "%s\n", file);
  buf = map->buffer;
  length = map->length;
  pos = 0;
  
  version = NEXTINT();
  if(version != VERSION) {
    fprintf(stderr, "link: \"%s\" has wrong version number.\n", file);
    if(!warning) {
      fprintf(stderr, "link: do not forget \"-f\".\n");
      warning = 1;
    }
    unmapFile(map);
    return 0;
  }

  globalOffset = NEXTINT();
  globalLength = NEXTINT();
  staticOffset = NEXTINT();
  staticLength = NEXTINT();
  stringsOffset = NEXTINT();
  stringsLength = NEXTINT();
  NEXTINT(); /* hashOffset */
  NEXTINT(); /* hashLength */
  dynamicOffset = NEXTINT();
  dynamicLength = NEXTINT();

  pos = globalOffset;
  globalSection(globalOffset + globalLength);

  pos = stringsOffset;
  stringsSection(stringsOffset + stringsLength);

  pos = staticOffset;
  staticSection(staticOffset + staticLength);

  pos = dynamicOffset;
  dynamicSection(dynamicOffset + dynamicLength);

  unmapFile(map);
  return 1;
}
