/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "list.h"
#include "bigarray.h"
#include "hashtable.h"
#include "registry.h"

struct local {
  int offset;
  struct variable *variable;
};

struct list *globalsList = 0;
struct list *stringsList = 0;

static struct bigarray *vars = 0;
static struct bigarray *deps = 0;
static struct bigarray *strings = 0;
static struct bigarray *locals = 0;

static struct hashtable *hashLocals = 0;
static struct hashtable *hashGlobals = 0;
static struct hashtable *hashStrings = 0;
static struct hashtable *hashtable = 0;

void initializeRegistry(void) {
  vars = createArray(sizeof(struct variable));
  deps = createArray(sizeof(struct dependency));
  strings = createArray(sizeof(struct string));
  locals = createArray(sizeof(struct local));
  
  hashGlobals = createTable(16);
  hashStrings = createTable(16);
  hashLocals = createTable(16);
  hashtable = createTable(16);
}

struct variable *newVariable(void) { return newElement(vars); }
struct dependency *newDependency(void) { return newElement(deps); }
struct variable *varOfIndex(int index) { return getElement(vars, index); }
struct dependency *depOfIndex(int index) { return getElement(deps, index); }
int varIndex(struct variable *var) { return getIndex(vars, var); }
int depIndex(struct dependency *dep) { return getIndex(deps, dep); }
int varCount(void) { return arraySize(vars); }
int depCount(void) { return arraySize(deps); }

void clearLocals(void) {
  int i;

  disposeElements(strings, 0);
  disposeElements(locals, 0);
  clearTable(hashLocals);
  clearTable(hashStrings);
  for(i=0; i<varCount(); i++) varOfIndex(i)->offset = 0;
}

static void *newLocal(int offset) {
  struct local *local = newElement(locals);
  
  local->offset = offset;
  return local;
}

static void *newLocalStr(int offset) {
  struct string *string = newElement(strings);

  string->offset = offset;
  return string;
}

static struct local *findLocal(int offset) {
  return findEntry1(hashLocals, offset, newLocal);
}

struct variable *findLocalVar(int offset) {
  struct local *local = findLocal(offset);

  if(!local->variable) {
    local->variable = newVariable();
    local->variable->offset = offset;
  }
  return local->variable;
}

struct string *findLocalStr(int offset) {
  return findEntry4(hashStrings, offset, newLocalStr);
}

static void *newGlobal(char *name) {
  struct variable *global = newVariable();
  struct list *tmp = globalsList;

  globalsList = newCons();
  globalsList->next = tmp;
  globalsList->elem = global;
  global->name = name;
  return global;
}

struct variable *findGlobal(int offset, char *name) {
  struct variable *global = findEntry2(hashGlobals, name, newGlobal);
  global->offset = offset;
  findLocal(offset)->variable = global;
  return global;
}

static void *newString(char *name) {
  struct string *string = newElement(strings);
  struct list *tmp = stringsList;

  string->name = name;
  stringsList = newCons();
  stringsList->next = tmp;
  stringsList->elem = string;
  return string;
}

int addString(char *name) {
  return ((struct string *)findEntry2(hashStrings, name, newString))->offset;
}

#define STRINGBUF 100000

static void *addStr(char *str) {
  static char *buf = 0;
  static int next = 0;
  char *ret;

  int len = strlen(str);
  
  if(len >= STRINGBUF) return STRDUP(str);
  
  if(!next || (strlen(str)+next >= STRINGBUF)) {
    buf = (char *)MALLOC(STRINGBUF);
    next = 0;
  }

  ret = buf+next;
  while(buf[next++]=*str++);
  return ret;
}

char *insertStr(char *str) { 
  return findEntry3(hashtable, str, addStr);
}

void stats(void) {
  fprintf(stderr, "Locals: %d\n", hashSize(hashLocals));
  fprintf(stderr, "Globals: %d\n", hashSize(hashGlobals));
  fprintf(stderr, "Strings: %d\n", hashSize(hashStrings));
  fprintf(stderr, "Strs: %d\n", hashSize(hashtable));

  fprintf(stderr, "Vars: %d\n", arraySize(vars) * sizeof(struct variable));
  fprintf(stderr, "Deps: %d\n", arraySize(deps) * sizeof(struct dependency));
  fprintf(stderr, "Strings: %d\n", arraySize(strings) * sizeof(struct string));

  fprintf(stderr, "Cons: %d\n", consSpaceUsage() * sizeof(struct list));
}
