/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include <sys/stat.h>
#include <unistd.h>

#include "fatal.h"
#include "bigarray.h"
#include "targets.h"

static struct bigarray *targets = 0;
static struct bigarray *nonTargets = 0;

void initializeTargets(void) {
  targets = createArray(sizeof(struct entry));
  nonTargets = createArray(sizeof(struct entry));
}

void resetTargets(void) { 
  struct entry *target;
  int i;
  
  for(i=0; i<targetCount(); i++) {
    target = targetOfIndex(i);
    if(target->name) free(target->name);
    if(target->ctype) free(target->ctype);
    if(target->file) free(target->file);
  }
  disposeElements(targets, 0);

  for(i=0; i<nonTargetCount(); i++) {
    target = nonTargetOfIndex(i);
    if(target->name) free(target->name);
    if(target->ctype) free(target->ctype);
    if(target->file) free(target->file);
  }
  disposeElements(nonTargets, 0);
}

struct entry *targetOfIndex(int index) { return getElement(targets, index); }
struct entry *nonTargetOfIndex(int index) { return getElement(nonTargets, index); }
int targetCount(void) { return arraySize(targets); }
int nonTargetCount(void) { return arraySize(nonTargets); }

struct entry *newTarget(char *name, char *ctype, char *file, int startLine, int endLine) {
  struct entry *target = newElement(targets);

  if(name) target->name = STRDUP(name);
  if(ctype) target->ctype = STRDUP(ctype);
  if(file) target->file = STRDUP(file);
  target->startLine = startLine;
  target->endLine = endLine;
  return target;
}

struct entry *newNonTarget(char *name, char *ctype, char *file, int startLine, int endLine) {
  struct entry *target = newElement(nonTargets);

  if(name) target->name = STRDUP(name);
  if(ctype) target->ctype = STRDUP(ctype);
  if(file) target->file = STRDUP(file);
  target->startLine = startLine;
  target->endLine = endLine;
  return target;
}

static int parse(char *buf, int size, struct entry *pattern) {
  int i = 0, j, startStr, lineNum;

  assert(pattern);

nameState:
  startStr = i;
  while(i < size) {
    switch(buf[i]) {
    case 0:
    case ' ':
    case '\t':
    case '\n':
    case ';':
      if(startStr < i) {
	buf[i] = 0;
	pattern->name = buf+startStr;
	pattern->ctype = 0;
	pattern->file = 0;
	pattern->startLine = 0;
	pattern->endLine = 0;
	return ++i;
      } else {
	if(!buf[i++]) return -1; /* EOF */
	goto nameState;
      }
      
    case '/':
      buf[i] = 0;
      if(startStr < i) {
	pattern->name = buf+startStr;
      } else {
	pattern->name = 0;
      }
      i++;
      goto typeState;
      
    case '<':
      buf[i] = 0;
      if(startStr < i) {
	pattern->name = buf+startStr;
      } else {
	pattern->name = 0;
      }
      i++;
      pattern->ctype = 0;
      goto locState;
      
    default:
      i++;
    }
  }
  return -1; /* EOF */

typeState:
  startStr = i;
  while(i < size) {
    switch(buf[i]) {
    case 0:
    case '\n':
    case ';':
      for(j=i; buf[j-1]==' '; j--);  /* skip trailing spaces */
      for(; buf[startStr] == ' '; startStr++);  /* skip leading spaces */
      buf[j] = 0;
      if(startStr < j) {
	pattern->ctype = buf+startStr;
      } else {
	pattern->ctype = 0;
      }
      pattern->file = 0;
      pattern->startLine = 0;
      pattern->endLine = 0;
      return ++i;
      
    case '<':
      for(j=i; buf[j-1]==' '; j--);  /* skip trailing spaces */
      for(; buf[startStr] == ' '; startStr++);  /* skip leading spaces */
      buf[j] = 0;
      pattern->ctype = buf+startStr;
      i++;
      goto locState;
      
    default:
      i++;
    }
  }
  goto patternError;

locState:
  startStr = i;
  while(i < size) {
    switch(buf[i]) {
    case 0:
    case '\n':
    case '>':
    case ';':
      for(j=i; buf[j-1]==' '; j--);  /* skip trailing spaces */
      for(; buf[startStr] == ' '; startStr++);  /* skip leading spaces */
      buf[j] = 0;
      pattern->file = buf+startStr;
      pattern->startLine = 0;
      pattern->endLine = 0;
      return ++i;
      
    case ':':
      for(j=i; buf[j-1]==' '; j--);  /* skip trailing spaces */
      for(; buf[startStr] == ' '; startStr++);  /* skip leading spaces */
      buf[j] = 0;
      pattern->file = buf+startStr;
      i++;
      goto startLineState;

    default:
      i++;
    }
  }
  goto patternError;

startLineState:
  lineNum = 0;
  while(i < size) {
    switch(buf[i]) {
    case 0:
    case '\n':
    case '>':
    case ';':
      pattern->startLine = lineNum;
      pattern->endLine = lineNum;
      return ++i;
      
    case '-':
      buf[i++] = 0;
      pattern->startLine = lineNum;
      goto endLineState;

    default: 
      if (buf[i] >= '0' && buf[i] <= '9') {
	lineNum = lineNum * 10 + (buf[i] - '0');
      } else {
	goto patternError;
      }
      i++;
    }
  }
  goto patternError;
  
endLineState:
  lineNum = 0;
  while(i < size) {
    switch(buf[i]) {
    case 0:
    case '\n':
    case '>':
    case ';':
      pattern->endLine = lineNum;
      return ++i;
      
    default: 
      if (buf[i] >= '0' && buf[i] <= '9') {
	lineNum = lineNum * 10 + (buf[i] - '0');
      } else {
	goto patternError;
      }
      i++;
    }
  }
  
patternError:
  return 0; /* error */
}

int parseTargets(char *buf, int size, int set) {
  struct entry target[1];
  int i = 0, ret, count = 0;
  
  while(i < size) {
    ret = parse(buf+i, size-i, target);
    if (!ret) { /* error */
      fprintf(stderr, "Error: incorrect pattern\n");
      break;
    }
    if(ret == -1) break; /* EOF */
    
    if(set == TARGETS) {
      newTarget(target->name, target->ctype, target->file, target->startLine, target->endLine);
    } else {
      newNonTarget(target->name, target->ctype, target->file, target->startLine, target->endLine);
    }
    i += ret;
    ++count;
  }
  return count;
}

int importTargets(char *file, int set) {
  FILE *f;
  struct stat statBuf;
  int size, newSize, count = 0;
  char *buf;
  
  if(stat(file, &statBuf)) {
    fprintf(stderr, "Error: can't find file %s\n", file);
    return 0;
  }

  size = statBuf.st_size;
  buf = (char *)MALLOC(size+1);
  
  f = fopen(file, "r");

  if(!f) {
    fprintf(stderr, "Error: can't find file %s\n", file);
    return 0;
  }

  newSize = fread(buf, 1, size, f);
  if (size != newSize) {
    fprintf(stderr, "Error: file %s corrupted: stat says file size is %d, but fread only reads %d bytes.\n", file, size, newSize);
  }
    
  if (newSize < size) {
    size = newSize;  /* continue as best we can */
  }
  
  buf[size] = '\n';  /* add new line at end */
  size++;
  
  count = parseTargets(buf, size, set);

  /* at the very end, all fall through and return */
  free(buf);
  fclose(f);
  return count;
}

int match(struct entry *filter, struct entry *entry) {
  assert(filter && entry);

  if((!filter->name || !strcmp(filter->name, entry->name)) &&
     (!filter->ctype || !strcmp(filter->ctype, entry->ctype)) &&
     (!filter->file || !strcmp(filter->file, entry->file)) &&
     (!filter->startLine || (filter->startLine == entry->startLine)) &&
     (!filter->endLine || (filter->endLine == entry->endLine))) {
    return 1;
  }
  return 0;
}

void fprintPattern(FILE *stream, struct entry *pattern) {
  assert(pattern);

  if(pattern->name) fprintf(stream, "%s", pattern->name);
  if(pattern->ctype) fprintf(stream, "/%s", pattern->ctype);
  if(pattern->file) {
    fprintf(stream, " <%s", pattern->file);
    if(pattern->startLine) fprintf(stream, ":%d", pattern->startLine);
    if(pattern->endLine != pattern->startLine) fprintf(stream, "-%d", pattern->endLine);
    fprintf(stream, ">");
  }
}

int parsePattern(char *string, struct entry *pattern) {
  if(pattern->name) free(pattern->name);
  if(pattern->ctype) free(pattern->ctype);
  if(pattern->file) free(pattern->file);

  if(parse(string, strlen(string) + 1, pattern) <= 0) {
    pattern->name = 0;
    pattern->ctype = 0;
    pattern->file = 0;
    pattern->startLine = 0;
    pattern->endLine = 0;
    return 0;
  } else {
    if(pattern->name) pattern->name = STRDUP(pattern->name);
    if(pattern->ctype) pattern->ctype = STRDUP(pattern->ctype);
    if(pattern->file) pattern->file = STRDUP(pattern->file);
    return 1;
  }
}

void copyPattern(struct entry *source, struct entry *pattern) {
  if(pattern->name) free(pattern->name);
  if(pattern->ctype) free(pattern->ctype);
  if(pattern->file) free(pattern->file);

  pattern->name = source->name?STRDUP(source->name):0;
  pattern->ctype = source->ctype?STRDUP(source->ctype):0;
  pattern->file = source->file?STRDUP(source->file):0;
  pattern->startLine = source->startLine;
  pattern->endLine = source->endLine;
}
