/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "list.h"
#include "storage.h"
#include "set.h"
#include "database.h"
#include "aliasing.h"

struct list *getProto(struct variable *var) {
  struct varOffsets offsets[1];
  struct list *vars, *list;
  
  if(var->prototype) return protoOfVar(var, 0);

  var->prototype = 1;
  getVarOffsets(var->offset, offsets);
  if(!offsets->prototype) return 0;

  getProtoInfos(offsets->prototype, vars = newCons());
  for(list=vars; list; list=list->next) {
    if(list->elem) list->elem = varAtOffset((int)list->elem);
  }
  return protoOfVar(var, vars);
}

void loadFunPtrDep(struct variable *var, int offset) {
  unsigned char dtype;
  
  if(offset) {
    dtype = getProtoInfos(offset, 0);
    if(isFunPtr(dtype)) newDep(offset, var, var, dtype);
  }
}

static int changeRequiringRecomputation = 0;

static void addLval(struct variable *var, struct variable *lval) {
  struct varInfos infos[1];

  getVarInfos(lval->offset, infos);
/*if(infos->name && infos->ctype && !strchr(infos->ctype, '*') && !strchr(infos->ctype, '(')) return;*/
  if(!var->initialLvals) var->initialLvals = emptySet();
  insertLval(lval, var->initialLvals);
  
  lval->initialPtrs = cons(var, lval->initialPtrs);
}

static void addConstraint(struct variable *var, struct set *lvalSet) {
  if(!var->lvals) var->lvals = emptySet();
  if(insertSet(lvalSet, var->lvals)) changeRequiringRecomputation = 1;
}

static unsigned char reverse(unsigned char dtype) {
  unsigned char tmp = dtype & (0xff ^ READDIRECT ^ WRITEDIRECT);
  
  if(dtype & READDIRECT) tmp = dtype | WRITEDIRECT;
  if(dtype & WRITEDIRECT) tmp = dtype | READDIRECT;

  return tmp;
}

static void loadVar(struct variable *var) {
  struct varOffsets offsets[1];
  struct depInfos infos[1];
  int offset, next;
  
  var->initialized = 1;
  getVarOffsets(var->offset, offsets);
  loadFunPtrDep(var, offsets->prototype);
  for(offset=offsets->forward; offset<offsets->backward; offset=next) {
    next = getDepInfos(offset, infos);
    if(infos->dtype != CALLSITE /*&& transferLvals(infos->dtype)*/) {
      /*
      if(isDirect(infos->dtype)) {
	if(!var->lvals) var->lvals = emptySet();
	addConstraint(varAtOffset(infos->var), var->lvals);
      } else {
      */
      newDep(offset, varAtOffset(infos->var), var, infos->dtype);
      /*}*/
    }
  }
  for(offset=offsets->backward; offset<offsets->end; offset=next) {
    next = getDepInfos(offset, infos);
    if(infos->dtype != CALLSITE && transferLvals(infos->dtype)) {
      newDep(offset, var, varAtOffset(infos->var), reverse(infos->dtype));
    }
  }
}

void track(struct variable *var) {
  struct list *ptrs;
  struct variable *ptr;

  if(var->track) return;
  changeRequiringRecomputation = 1;
  var->track = 1;
  if(!isInitialized(var)) loadVar(var);
  for(ptrs=var->initialPtrs; ptrs; ptrs=ptrs->next) {
    ptr = ptrs->elem;
    if(!ptr->lvals) ptr->lvals = emptySet();
    insertLval(var, ptr->lvals);
    track(ptr);
  }
}

void demand(struct variable *var) {
  if(var->demand) return;
  changeRequiringRecomputation = 1;
  var->demand = 1;
  if(!isInitialized(var)) loadVar(var);
  if(var->initialLvals) addConstraint(var, var->initialLvals);
  track(var);
}

static void recompute(struct dependency *dep);

static void recomputeFunPtr(struct variable *var) {
  struct list *lvals, *ptrProto, *funProto, *saveProto;
  struct variable *lval, *x, *y;
  struct dependency dep[1];
  int args;
  
  if(hasNoLvals(var)) return;
  saveProto = getProto(var);
  if(!saveProto) return;
  
  for(lvals=getElems(var->lvals); lvals; lvals=lvals->next) {
    lval = (struct variable *)lvals->elem;
    funProto = getProto(lval);
    if(!funProto) continue;

    ptrProto = saveProto;
    args = 0; /* first return value */
    
    while(ptrProto && funProto) {
      if(args) {
	x = ptrProto->elem;
	y = funProto->elem;
      } else {
	x = funProto->elem;
	y = ptrProto->elem;
      }

      if(x && y) {
	dep->offset = 0;
	dep->dtype = DEP | STRONG | DIRECT | READDIRECT | WRITEDIRECT | TRANSFERLVALS;
	dep->readVar = x;
	dep->writeVar = y;
	recompute(dep);
      }
      
      ptrProto = ptrProto->next;
      funProto = funProto->next;
      args = 1; /* then arguments */
    }
  }
}

static void recompute(struct dependency *dep) {
  struct list *lvals;
  struct variable *readVar, *writeVar, *lval;

  if(isFunPtr(dep->dtype)) {
    recomputeFunPtr(dep->writeVar);
    return;
  }

  readVar = dep->readVar;
  writeVar = dep->writeVar;
  
  if(isDirect(dep->dtype)) {
    if(writeVar->demand) demand(readVar);
    if(readVar->lvals) {
      addConstraint(writeVar, readVar->lvals);
      track(writeVar);
    }
    return;
  }

  if(isReadPtr(dep->dtype)) {
    if(writeVar->demand) demand(readVar);
    if(writeVar->demand && readVar->lvals) {
      for(lvals=getElems(readVar->lvals); lvals; lvals=lvals->next) demand(lvals->elem);
    }
    if(readVar->lvals) {
      addConstraint(writeVar, getElemElems(readVar->lvals));
      track(writeVar);
    }
    return;
  }
  
  if(isWritePtr(dep->dtype)) {
    if(writeVar->lvals) demand(readVar); /* change? */
    if(readVar->lvals) {
      demand(writeVar);
      for(lvals=getElems(writeVar->lvals); lvals; lvals=lvals->next) {
	lval = lvals->elem;
	addConstraint(lval, readVar->lvals);
	track(lval);
      }
    }
    return;
  }
}

/*
static void recompute(struct dependency *dep) {
  struct list *lvals, *lsets;
  struct variable *readVar, *writeVar;
  struct set *lvalsSoFar;

  if(isFunPtr(dep->dtype)) {
    recomputeFunPtr(dep->writeVar);
    return;
  }
  
  readVar = dep->readVar;
  writeVar = dep->writeVar;
  
  if(hasNoLvals(readVar)) return;
  
  if(isReadPtr(dep->dtype)) {
    lvalsSoFar = getElemElems(readVar->lvals);
    for(lvals = todo; lvals; lvals=lvals->next) {
      if(!isInitialized((struct variable *)lvals->elem)) {
	loadVar((struct variable *)lvals->elem);
	changeRequiringRecomputation = 1;
      }
    }
    freeL(todo);
    todo = 0;
    if(isEmptySet(lvalsSoFar)) return;
  } else { 
    lvalsSoFar = readVar->lvals;
  }
  
  if(isWritePtr(dep->dtype)) {
    for(lsets = getSets(writeVar->lvals); lsets; lsets=lsets->next) {
      if(insertSet(lvalsSoFar, lsets->elem)) changeRequiringRecomputation = 1;
    }
    for(lvals = todo; lvals; lvals=lvals->next) {
      if(!isInitialized((struct variable *)lvals->elem)) {
	loadVar((struct variable *)lvals->elem);
	changeRequiringRecomputation = 1;
      }
    }
    freeL(todo);
    todo = 0;
  } else {
    addConstraint(writeVar, lvalsSoFar);
  }
}
*/
void loadStaticDeps(void) {
  struct staticInfos infos[1];
  int i;

  i = getStaticInfos(0, infos);
  while(i) {
    addLval(varAtOffset(infos->var), varAtOffset(infos->lval));
    i = getStaticInfos(i, infos);
  }
  for(i=0; i<funPtrCount(); i++) demand(varAtOffset(funPtrOfIndex(i)));
}

void aliasing(void) {
  int i;

  for(changeRequiringRecomputation = 1; changeRequiringRecomputation; ) {
    changeRequiringRecomputation = 0;
    currentIterationIndex++;
    if(debug) fprintf(stderr, ".");
    /*    malloc_stats();*/
    gcLvalLists();    
    for(i=0; i<depCount(); i++) recompute(depOfIndex(i));
  }
}
