/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "list.h"
#include "storage.h"
#include "set.h"
#include "database.h"
#include "aliasing.h"
#include "targets.h"
#include "analyse.h"

static int changeRequiringRecomputation = 0;
static int propagate = 0;
static int direction = FORWARD;

#define combineDtype(dt1, dt2) ((dt1) & (dt2))
#define isBetterDtype(dt1, dt2) (((dt1) & CMPBITS) > ((dt2) & CMPBITS))
#define isSameDtype(dt1, dt2) (((dt1) & CMPBITS) == ((dt2) & CMPBITS))
#define isBetterChain(currLen, len, currDt, dt) (isBetterDtype((dt), (currDt)) || (isSameDtype((dt), (currDt)) && ((len) < (currLen))))

#define FUZZ 10

static void matchTarget(int offset, struct varInfos *infos, struct entry *target) {
  struct variable *var;
  struct entry entry[1];
  
  for(entry->name = infos->name; *entry->name && *entry->name != '.'; entry->name++);
  if(*entry->name) entry->name++; else entry->name = infos->name;

  entry->ctype = infos->ctype;
  entry->file = infos->loc.file;
  entry->startLine = infos->loc.startLine;
  entry->endLine = infos->loc.endLine;

  if(match(target, entry)) {
    var = varAtOffset(offset);
    var->chain = newChain();
    var->chain->writeVar = var;
    var->chain->dtype = DEP | STRONG | DIRECT | READDIRECT | WRITEDIRECT | TRANSFERLVALS;
  }
}

void loadTargets(void) {
  struct varInfos infos[1];
  int i, offset;
  
  for(i=0; i<targetCount(); i++) {
    offset = findTarget(targetOfIndex(i)->name, infos);
    while(offset) {
      matchTarget(offset, infos, targetOfIndex(i));
      offset = findTarget(0, infos);
    }
  }
}

static void matchNonTarget(struct variable *var) {
  struct varInfos infos[1];
  struct entry entry[1];
  int i;
  
  if(!nonTargetCount()) return;

  getVarInfos(var->offset, infos);
  
  for(entry->name = infos->name; *entry->name && *entry->name != '.'; entry->name++);
  if(*entry->name) entry->name++; else entry->name = infos->name;
  
  entry->ctype = infos->ctype;
  entry->file = infos->loc.file;
  entry->startLine = infos->loc.startLine;
  entry->endLine = infos->loc.endLine;

  for(i=0; i<nonTargetCount(); i++) {
    if(match(nonTargetOfIndex(i), entry)) {
      var->nonTarget = 1;
      return;
    }
  }
}

static void loadVar(struct variable *var);

static void updateChain(struct variable *writeVar, struct variable *readVar, int offset, unsigned char dtype, int length) {
  int change = 0;
  
  if(isNonTargetVar(writeVar)) return;

  if(!hasDep(writeVar)) {
    writeVar->chain = newChain();
    change = 1;
  } else {
    if(isBetterChain(writeVar->chain->length, length, writeVar->chain->dtype, dtype)) change = 1;
  }

  if(change) {
    writeVar->chain->offset = offset;
    writeVar->chain->writeVar = writeVar;
    writeVar->chain->readVar = readVar;
    writeVar->chain->dtype = dtype;
    writeVar->chain->length = length;
    changeRequiringRecomputation = 1;
    if (propagate || !isInitialized(writeVar)) {
      loadVar(writeVar);
      writeVar->chain->changed = 0;
    } else {
      writeVar->chain->changed = 1;
    }
  }
}

static void updateLvalsChain(struct variable *var, struct chain *chain) {
  if(isNonTargetVar(var)) return;
  
  if(!hasLvalsDep(var)) {
    var->lvalsChain = chain;
  } else {
    if(isBetterChain(var->lvalsChain->length, chain->length, var->lvalsChain->dtype, chain->dtype)) {
      var->lvalsChain = chain;
    }
  }
}

static void recomputeFunPtr(struct variable *var) {
  struct list *lvals, *ptrProto, *funProto;
  struct variable *lval, *x, *y;
  int args;
  
  if(hasNoLvals(var)) return;
  ptrProto = getProto(var);
  if(!ptrProto) return;
  
  for(lvals=getElems(var->lvals); lvals; lvals=lvals->next) {
    lval = (struct variable *)lvals->elem;
    funProto = getProto(lval);
    if(!funProto) continue;

    args = 0; /* first return value */
    
    while(ptrProto && funProto) {
      if(args) {
	x = ptrProto->elem;
	y = funProto->elem;
      } else {
	x = funProto->elem;
	y = ptrProto->elem;
      }

      if(x && y) {
	if(direction != BACKWARD) {
	  if(hasDep(x)) updateChain(y, x->chain->readVar, x->chain->offset, x->chain->dtype, x->chain->length);
	}
	
	if(direction != FORWARD) {
	  if(hasDep(y)) updateChain(x, y->chain->readVar, y->chain->offset, y->chain->dtype, y->chain->length);
	}
      }
      ptrProto = ptrProto->next;
      funProto = funProto->next;
      args = 1; /* then arguments */
    }
  }
}

static void recompute(struct dependency *dep) {
  struct variable *lval, *readVar, *writeVar;
  struct list *lvals;
  struct chain *chain = 0;

  readVar = dep->readVar;
  writeVar = dep->writeVar;
  
  if(isFunPtr(dep->dtype)) {
    recomputeFunPtr(dep->writeVar);
    return;
  }
  
  if(isReadPtr(dep->dtype)) {
    for(lvals = getElems(readVar->lvals); lvals; lvals=lvals->next) {
      lval = (struct variable *)lvals->elem;
      if(hasDep(lval)) updateLvalsChain(readVar, lval->chain);
    }
    chain = readVar->lvalsChain;
  } else { 
    chain = readVar->chain;
  }
  
  if(!chain) return;
  
  if(isWritePtr(dep->dtype)) {
    for(lvals = getElems(writeVar->lvals); lvals; lvals=lvals->next) {
      lval = (struct variable *)lvals->elem;
      updateChain(lval, readVar, dep->offset, combineDtype(dep->dtype, chain->dtype), chain->length+1);
    }
  } else {
    updateChain(writeVar, readVar, dep->offset, combineDtype(dep->dtype, chain->dtype), chain->length+1);
  }
}

static void loadVar(struct variable *var) {
  struct varOffsets offsets[1];
  struct depInfos infos[1];
  int offset, next, again = isInitialized(var), first, last;
  
  getVarOffsets(var->offset, offsets);
  if(!again) {
    var->initialized = 1;
    matchNonTarget(var);
    if(isNonTargetVar(var)) return; 
    loadFunPtrDep(var, offsets->prototype);
  }

  switch(direction) {
  case FORWARD: first = offsets->forward; last = offsets->backward; break;
  case BACKWARD: first = offsets->backward; last = offsets->end; break;
  case UNIFICATION: first = offsets->forward; last = offsets->end;
  }

  for(offset=first; offset<last; offset=next) {
    next = getDepInfos(offset, infos);
    if(isDirect(infos->dtype)) {
      if(hasDep(var)) updateChain(varAtOffset(infos->var), var, offset, combineDtype(infos->dtype, var->chain->dtype), var->chain->length+1);
    } else {
      if(!again) newDep(offset, varAtOffset(infos->var), var, infos->dtype);
    }
  }
}

void analyse(int mode) {
  struct variable *var;
  struct chain *chain;
  int i;

  propagate = 0;
  direction = mode;

  loadTargets();

  for(i=0; i<varCount(); i++) {
    var = varOfIndex(i);
    if(!isInitialized(var) && (isTargetVar(var) || !hasNoLvals(var))) loadVar(var);
  }

  for(changeRequiringRecomputation = 1; changeRequiringRecomputation; ) {
    changeRequiringRecomputation = 0;
    if(debug) fprintf(stderr, ".");
    for(i=0; i<chainCount(); i++) {
      chain = chainOfIndex(i);
      
      if(chain->changed) {
	chain->changed = 0;
	loadVar(chain->writeVar);
      }
    }
  }

  propagate = 1;

  for(changeRequiringRecomputation = 1; changeRequiringRecomputation; ) {
    changeRequiringRecomputation = 0;
    currentIterationIndex++;
    if(debug) fprintf(stderr, ".");
    gcLvalLists();
    for(i=0; i<depCount(); i++) recompute(depOfIndex(i));
  }
}
