/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "list.h"
#include "hashtable.h"
#include "bigarray.h"
#include "storage.h"

struct prototype {
  int offset;
  struct list *list;
};

static struct bigarray *vars = 0;
static struct bigarray *deps = 0;
static struct bigarray *chains = 0;
static struct bigarray *protos = 0;

static struct hashtable *hashtable = 0;

static int state = 0;

void initialize(void) {
  vars = createArray(sizeof(struct variable));
  deps = createArray(sizeof(struct dependency));
  chains = createArray(sizeof(struct chain));
  protos = createArray(sizeof(struct prototype));

  hashtable = createTable(16);
}

static void disposePrototypes(void) {
  struct list *list, *l;
  struct prototype *proto;
  int i;

  for(i=0; i<arraySize(protos); i++) {
    proto = getElement(protos, i);
    for(list=proto->list; list; list=l) {
      l = list->next;
      freeCons(list);
    }
  }
  
  disposeElements(protos, 0);
}

void saveState(void) {
  int i;

  state = varCount();
  disposeElements(deps, 0);
  for(i=0; i<state; i++) varOfIndex(i)->initialized = 0;
}

static struct variable *currentVar;

static void *this(int offset) { offset = offset; return currentVar; }

void restoreState(void) {
  int i;
  
  disposeElements(deps, 0);
  disposeElements(chains, 0);
  disposePrototypes();
  disposeElements(vars, state);

  clearTable(hashtable);
  
  for(i=0; i<varCount(); i++) {
    currentVar = varOfIndex(i);
    findEntry(hashtable, currentVar->offset, this);
    currentVar->chain = 0;
    currentVar->lvalsChain = 0;
    currentVar->initialized = 0;
    currentVar->prototype = 0;
    currentVar->nonTarget = 0;
  }
}

static void *addVar(int offset) {
  struct variable *var = newElement(vars);

  var->offset = offset;
  return var;
}

struct variable *varAtOffset(int offset) { return findEntry(hashtable, offset, addVar); }

struct dependency *newDep(int offset, struct variable *writeVar, struct variable *readVar, unsigned char dtype) {
  struct dependency *dep = newElement(deps);

  dep->offset = offset;
  dep->writeVar = writeVar;
  dep->readVar = readVar;
  dep->dtype = dtype;
  return dep;
}

struct chain *newChain(void) { return newElement(chains); }
struct variable *varOfIndex(int index) { return getElement(vars, index); }
struct dependency *depOfIndex(int index) { return getElement(deps, index); }
struct chain *chainOfIndex(int index) { return getElement(chains, index); }
int varIndex(struct variable *var) { return getIndex(vars, var); }
int depIndex(struct dependency *dep) { return getIndex(deps, dep); }
int chainIndex(struct chain *chain) { return getIndex(chains, chain); }
int varCount(void) { return arraySize(vars); }
int depCount(void) { return arraySize(deps); }
int chainCount(void) { return arraySize(chains); }

static struct list *list;

static void *addProto(int offset) {
  struct prototype *prototype = newElement(protos);

  prototype->offset = offset;
  prototype->list = list;
  return prototype;
}

struct list *protoOfVar(struct variable *var, struct list *proto) {
  struct prototype *prototype;
  
  if(proto) {
    list = proto;
    assert(!findEntry(hashtable, var->offset+1, 0));
    prototype = findEntry(hashtable, var->offset+1, addProto);
    return prototype->list;
  } else {
    prototype = findEntry(hashtable, var->offset+1, 0);
    return prototype?prototype->list:0;
  }
}
