(**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************)

structure GDAFile =  
  struct
    
    val dependencyBit = 0wx1
    val strongBit = 0wx2
    val directBit = 0wx4
    val readDirectBit = 0wx8
    val writeDirectBit = 0wx10
    val transferLvalsBit = 0wx20
    val nonPtrType = 0wx80
    val ptrType = 0wx81
      
    val varCount = ref 0
    val anonVars = ref 0
    val progVars = ref 0
    val dynamicDependencyCount = ref 0
      
    val initialPhase = ref true
      
    val globalOffset = ref 0
    val endglobalOffset = ref 0
    val staticOffset = ref 0
    val endstaticOffset = ref 0
    val stringsOffset = ref 0
    val endstringsOffset = ref 0
    val hashOffset = ref 0
    val endhashOffset = ref 0
    val dynamicOffset = ref 0
    val enddynamicOffset = ref 0
    val endOfFileOffset = ref 0
    val constantsOffset = ref 0
    val endconstantsOffset = ref 0
    val hashConstantsOffset = ref 0
    val endhashConstantsOffset = ref 0
      
    open Types
    open Output
      
    datatype varInfo = 
      Offset of int
    | Offsets of (int * int * int * int)
      
    type varOffsets = {offset: int, offsets: (int * int * int * int)}
      
    exception varNotFound
    
    val varsTable =
      UidMap.mkTable (100, varNotFound): varOffsets UidMap.hash_table
      
    fun insertVar (v, info) =
      let
	val uid = varToPid v
	val {offset, offsets} = UidMap.lookup varsTable uid
	  handle varNotFound => {offset=0, offsets=(0,0,0,0)}
      in
	case info of
	  Offset i =>
	    UidMap.insert varsTable (uid, {offset=i, offsets=offsets})
	| Offsets s => 
	    UidMap.insert varsTable (uid, {offset=offset, offsets=s})
      end
    
    fun findVarOffset v = 
      (SOME (#offset (UidMap.lookup varsTable (varToPid v)))
       handle varNotFound => NONE)
      
    fun findVarOffsets v =
      (SOME (#offsets (UidMap.lookup varsTable (varToPid v)))
       handle varNotFound => NONE)
      
    val warning = print
      
    exception fatalGDAFile
    
    fun fatal s = (print s; raise fatalGDAFile)
      
    val debug = ref false
    val pids = ref false

    fun debugVarName v =
      case v of 
	V{pid, name, ...} =>
	  (name ^ "[" ^ (Int.toString pid) ^ "]" )
      | A{pid, ...} =>
	  ("@tmp[" ^ (Int.toString pid) ^ "]")
      | F{pid, arg, var, ...} =>
	  ((debugVarName var) ^ "#" ^ (Int.toString arg) ^ "[" ^
	   (Int.toString pid) ^ "]")
      | R{pid, var, ...} =>
	  ((debugVarName var) ^ "()[" ^ (Int.toString pid) ^ "]")
	
    structure StringMap =
      HashTableFn(struct
		    type hash_key = string
		    fun hashVal s = 
		      let
			fun f nil = 0wx0
			  | f (c :: cl) =
			  Word.andb (((Word.fromInt (Char.ord c)) * 0wx57) +
				     (f cl), 0wxffffff)
		      in
			f (String.explode s)
		      end
		    fun sameKey (k1, k2) = (k1 = k2)
		  end)
	 
    exception stringNotFound
    
    val stringsTable =
      StringMap.mkTable (100, stringNotFound): int StringMap.hash_table
      
    fun insertString s =
      StringMap.lookup stringsTable s
      handle stringNotFound => (StringMap.insert stringsTable (s, 0); 0)
	
    fun insertStringOffset (s, offset) =
      StringMap.insert stringsTable (s, offset)
      
    fun writeLoc loc =
      let val (stringsOffset, i1, i2) = 
	case loc of
	  SourceMap.LOC {srcFile, beginLine, endLine, ...} =>
	    (srcFile, beginLine, endLine)
	| SourceMap.UNKNOWN => ("Unknown", ~1, ~1) 
      in
	writeInt (insertString stringsOffset);
	writeInt i1;
	writeInt i2
      end
    
    fun getOffset v =
      (case findVarOffset v of
	 SOME i => i
       | NONE =>
	   (warning ("Can't get var offset: " ^ (debugVarName v) ^ "\n"); 0))
      
    fun writeVar v =
      if !initialPhase then writeInt 0 else writeInt (getOffset v)
      
    fun writeStrings () =
      let fun f (str, _) = (insertStringOffset (str, !pos);
			    writeString str)
      in
	StringMap.appi f stringsTable
      end
      
    fun writeStaticDep (Addrof {write, lval, loc}) = 
      (writeChar #"&";
       writeVar write;
       writeVar (V lval);
       writeLoc loc;
       if !debug andalso !initialPhase then 
	 print ("STATIC: " ^ (debugVarName write) ^ " =& " ^
		(debugVarName (V lval)) ^ "\n")
	 else ())
      
    fun writeStaticDeps staticDeps = List.app writeStaticDep staticDeps
      
    fun writeDep reversed (SimpleDep {write, writeDirect, read, readDirect,
				      loc, dcat, lstr, rstr}) = 
      let
	fun depCategoryToBits STRONG_LVAL =
	  Word.orb (dependencyBit, Word.orb (strongBit, transferLvalsBit))
	  | depCategoryToBits STRONG = Word.orb (dependencyBit, strongBit)
	  | depCategoryToBits WEAK = dependencyBit
	  | depCategoryToBits NOTHING =
	  (warning "Internal Error: depCategoryToBits called with NOTHING\n";
	   0wx0)  (* should never happen e.g. see Unop and Binop *)
	  val db_0_1_5 = depCategoryToBits dcat
	  val db2 = if readDirect andalso writeDirect then directBit else 0wx0
	  val db3 = if readDirect then readDirectBit else 0wx0
	  val db4 = if writeDirect then writeDirectBit else 0wx0
	  val depTypeBits = 
	    Word.orb (db_0_1_5, Word.orb (db2, Word.orb(db3, db4)))
      in
	writeChar (Char.chr (Word.toInt (Word.andb (depTypeBits, 0wxff))));
	writeVar (if reversed then read else write);
	writeLoc loc;
	writeString lstr;
	writeString rstr;
	if reversed andalso !debug andalso !initialPhase then
	  print ("SIMPLE: " ^ (if writeDirect then "" else "*") ^
		 (debugVarName write) ^ " = " ^ 
		 (if readDirect then "" else "*") ^
		 (debugVarName read) ^ "\n")
	else ()
      end
    
    | writeDep reversed (CallSite {caller, calle, loc}) = 
      (writeChar (Char.chr (Word.toInt (Word.andb (0wx82, 0wxff))));
       writeVar (if reversed then (V caller) else calle);
       writeLoc loc;
       if reversed andalso !debug andalso !initialPhase then
	   print ("CALLSITE: " ^
		  (debugVarName (V caller)) ^ " calls " ^ 
		  (debugVarName calle) ^ "\n")
       else ())

    fun setOffset x =
      if !initialPhase then x := !pos
      else if (!x) <> (!pos) then
	fatal "Internal Error: offset is incorrect\n" else ()
	  
    fun writeGlobals wil =
      let
	fun test ({used=true, var=v as
		   V {global=true, name, ...}, ...}: varInfos) =
	  (writeString name; writeVar v)
	  | test {var=v as R {var=V {name, ...}, ...}, ...} =
	  (writeString (name ^ "()"); writeVar v)
	  | test {var=v as F {var=V {name, ...}, arg, ...}, ...} =
	  (writeString (name ^ "#" ^ (Int.toString arg)); writeVar v)
	  | test _ = ()
      in
	List.app test wil
      end
    
    (* fix the hash function f *)
    fun writeConstants consts =
      let
	val hashSize = (List.length consts) * 2
	val cstHash = Array.array (hashSize, 0)
	  
	fun hashCst cst =
	  let
	    val cList = explode cst
	    fun f (hash, nil) = hash
	      | f (hash, c :: cl) =
	      f (((hash * 67) + (ord c)) mod hashSize, cl)
	    val hash = f (0, cList)
	  in
	    hash
	  end
	
	fun addCst (hash, offset) =
	  if Array.sub (cstHash, hash) = 0 then
	    Array.update (cstHash, hash, offset)
	  else
	    addCst ((hash + 1) mod hashSize, offset)
	  
	fun writeConstant (cst, list) =
	  (addCst (hashCst cst, !pos);
	   writeInt (size cst + 5 + 12 * (length list));
	   writeString cst;
	   List.app writeLoc list)
      in
	List.app writeConstant consts;
	cstHash
      end
    
    fun writeProto _ NONE = ()
      | writeProto var (SOME {ptr, ret, args}) =
      (((if ptr then 
	   writeChar (Char.chr (Word.toInt (Word.andb (ptrType, 0wxff))))
	 else
	   writeChar (Char.chr (Word.toInt (Word.andb (nonPtrType, 0wxff)))));
	(case ret of NONE => writeInt 0 | SOME r => writeVar r));
       writeInt (length args);
       app writeVar args;
       if !debug andalso !initialPhase then
	 print ("PROTO : " ^ (debugVarName var) ^ " | " ^ (Bool.toString ptr) ^ "\n")
       else ())
	 
    (* fix *)
    fun writeDeps ttab wil = 
      let val _ = (varCount := (List.length wil))
	(* initialize name hashing scheme for variables *)
	val hashSize =
	  let
	    val count = 2 * List.length wil
	    fun calc size =
	      if size >= count then size
	      else calc (2 * size)
	  in
	    calc 1
	  end
	val varHash = Array.array (hashSize, 0)
	fun hashStr str =
	  let val cList = explode str
	    fun f (hash, nil) = hash
	      | f (hash, #"." :: cl) = f(0, cl)
	      (* ignore everything before a '.' *)
	      | f (hash, c :: cl) =
	      f (((hash * 67) + (ord c)) mod hashSize, cl)
	  in
	    f (0, cList)
	  end
	
	(* WARNING: OBSCURE CODE! *)
	fun addVar(str, offset) =
	  let
	    fun findSpot hash =
	      if Array.sub (varHash, hash) = 0 then 
		Array.update (varHash, hash, offset)
	      else findSpot ((hash + 17) mod hashSize)
	  in
	    if str = "" then () else findSpot (hashStr str)
	  end
	
	fun f ({used=false, var=v, ...}:varInfos) = varCount := !varCount - 1
	  | f {dep=depl, rev=depl2, var=v, context, proto, ...} =
	  let val offset = !pos in
	    if !initialPhase then insertVar (v, Offset offset)
	    else
	      (case findVarOffset v of
		 SOME offset' => (if offset <> offset' then
				    fatal ("Offset indexing is invalid"
					   ^ (Int.toString (offset - offset'))
					   ^ "\n")
				  else ())
	       | NONE => fatal "Can't find var offset\n");
	      let val (commonOffset, depOffset, revOffset, endOffset) =
		if !initialPhase then (0, 0, 0, 0)
		else (case findVarOffsets v of
			SOME n => n
		      | NONE => fatal "Can't find var offsets\n")
	      in
		writeInt (commonOffset);
		writeInt (depOffset);
		writeInt (revOffset);
		writeInt (endOffset)
	      end;
	      (case v of
		 A _ => anonVars := !anonVars + 1
	       | V _ => progVars := !progVars + 1
	       | _ => ());
	    let
	      val commonOffset = !pos
	      val (name, loc, pid, tp) =
		case v of
		  V {name, loc, pid, tp, ...} =>
		    (name, loc, pid, ctypeToString ttab tp)
		| A {loc, pid, ...} =>
		    ("", loc, pid, "")
		| F {pid, arg, var=V {name, loc, ...}, ...} =>
		    (name ^ "#" ^ (Int.toString arg), loc, pid, "")
		| F {pid, arg, var=A {loc, ...}, ...} =>
		    ("", loc, pid, "")
		| F {pid, arg, ...} =>
		    ("", SourceMap.UNKNOWN, pid, "")
		| R {pid, var=V {name, loc, tp, ...}, ...} =>
		    (name ^ "()", loc, pid, ctypeToString ttab tp)
		| R {pid, var=A {loc, ...}, ...} =>
		    ("", loc, pid, "")
		| R {pid, ...} =>
		    ("", SourceMap.UNKNOWN, pid, "")
	    in
	      if !pids then writeString (Int.toString pid) else writeString name;
	      addVar (name, getOffset v);
	      writeLoc loc;
	      writeString tp;
	      (if !initialPhase then writeInt 0
	       else
		 (case context of
		    SOME function => (case findVarOffset (V function) of
					SOME offs => writeInt offs
				      | NONE => writeInt 0)
		  | NONE => writeInt 0));

	      let val depOffset = !pos in
		dynamicDependencyCount := ((List.length depl) +
					   (!dynamicDependencyCount));
		writeProto v proto;
		List.app (writeDep false) depl;
		let val revOffset = (!pos) in 
		  List.app (writeDep true) depl2;
		  insertVar(v, Offsets (commonOffset, depOffset,
					revOffset, !pos))
		end
	      end
	    end
	  end
      in
	List.app f wil;
	varHash  (* return hash array *)
      end
    
    fun writeAll ttab staticDeps wil consts =
      (writeInt 3; (* version *)
       writeInt (!globalOffset); 
       writeInt (!endglobalOffset - !globalOffset); 
       
       writeInt (!staticOffset); 
       writeInt (!endstaticOffset - !staticOffset); 
       
       writeInt (!stringsOffset); 
       writeInt (!endstringsOffset - !stringsOffset); 
       
       writeInt (!hashOffset); 
       writeInt (!endhashOffset - !hashOffset); 
       
       writeInt (!dynamicOffset); 
       writeInt (!enddynamicOffset - !dynamicOffset); 
       
       writeInt (!hashConstantsOffset);
       writeInt (!endhashConstantsOffset - !hashConstantsOffset);
       
       writeInt (!constantsOffset);
       writeInt (!endconstantsOffset - !constantsOffset);
       
       setOffset globalOffset;
       if !pids then () else writeGlobals wil;
       setOffset endglobalOffset;
       
       setOffset staticOffset;
       writeStaticDeps staticDeps;
       setOffset endstaticOffset;
       
       setOffset dynamicOffset;
       let
	 val hashArray = writeDeps ttab wil
	 val _ = setOffset enddynamicOffset
	 val size = Array.length hashArray
	 fun writeArray i =
	   if i < size then (writeInt (Array.sub (hashArray, i));
			     writeArray (i+1)) else ()
       in
	 setOffset hashOffset;
	 writeArray 0;
	 setOffset endhashOffset
       end;
       
       setOffset constantsOffset;
       let
	 val hashArray = writeConstants consts
	 val _ = setOffset endconstantsOffset;
	 val size = Array.length hashArray
	 fun writeArray i =
	   if i < size then (writeInt (Array.sub (hashArray, i));
			     writeArray (i+1)) else ()
       in
	 setOffset hashConstantsOffset;
	 writeArray 0;
	 setOffset endhashConstantsOffset
       end;
       
       setOffset stringsOffset;
       writeStrings ();
       setOffset endstringsOffset;
       
       setOffset endOfFileOffset)
      
    fun writeGDAFile stream ttab deb p staticDeps wil consts =
      (debug := deb;
       pids := p;
       StringMap.clear stringsTable;
       UidMap.clear varsTable;
       initialPhase := true;
       pos := 0;
       outStream := NONE;
       writeAll ttab staticDeps wil consts;
       
       varCount := 0;
       progVars := 0;
       anonVars := 0;
       dynamicDependencyCount := 0;
       
       initialPhase := false;
       pos := 0;
       outStream := SOME stream;
       writeAll ttab staticDeps wil consts;
       
       print ("Variables " ^ (Int.toString (!varCount)) ^ "\n");
       print ("Program Variables " ^ (Int.toString (!progVars)) ^ "\n");
       print ("Anonymous Variables " ^ (Int.toString (!anonVars)) ^ "\n");
       print ("Dynamic Dependencies " ^
	      (Int.toString (!dynamicDependencyCount)) ^ "\n"))
  end
