(**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************)

structure Track =
  struct
    open Ast
    open Types
      
    val warning = print
      
    type adornment = Aid.uid
      
    datatype varInfo =
      Dep of dynDep
    | Rev of dynDep
    | Context of vid
    | Proto of funInfos
    | Used
      
    datatype rval =
      Nothing
    | Const of string
    | Var of var
    | AddrOfId of {vid: vid, adornment: adornment, loc: loc}
    | DeRef of {lstr: string, rval: rval, rstr: string,
		adornment: adornment, loc: loc}
    | Op1 of {lstr: string, rval: rval, rstr: string, dcat: depCategory}
    | Op2 of {rval1: rval, dcat1: depCategory, str: string,
	      rval2: rval, dcat2: depCategory}
    | Join of {rval1: rval, rval2: rval}
      
    structure ConstMap =
      HashTableFn (struct
		     type hash_key = string
		     fun hashVal s =
			 let
			     fun f nil = 0wx0
			       | f (c::cl) =
				 Word.andb (((Word.fromInt (Char.ord c)) * 0wx57) +
					    (f cl), 0wxffffff) in 
			   f (String.explode s)
			 end
		     fun sameKey (k1, k2) = (k1 = k2)
		   end)

    fun travAst (tidtab, aidtab) {mallocs, noFields, alwaysTrackLvals,
				  commonSizeFields, commonTypeFields, noStrings} edl =
	let
	    
	    (* Malloc *)
	    val checkForMallocs = not (mallocs = nil)
	      
	    (* Fields *)
	    val anonHashSize = IntInf.fromInt 0x3fffffff
	    val anonTidtab = Tidtab.uidtab(): string Tidtab.uidtab
	    val memberPidtab = Pidtab.uidtab(): string Pidtab.uidtab
	      
	    (* possible delete *)
	    fun asciFy x = 
		let
		    val charSize = 62
		    fun mkChar x (* should be between 0 and 61 *) = 
			if x < 0 then #"0" 
			else if x < 10 then chr((ord #"0") + x)
			     else if x < 36 then chr((ord #"a") + (x-10))
				  else if x < 62 then chr((ord #"A") + (x-36))
				       else #"0"
		    val c0 = mkChar(x mod charSize)
		    val t1 = x div charSize
		    val c1 = mkChar(t1 mod charSize)
		    val t2 = t1 div charSize
		    val c2 = mkChar(t2 mod charSize)
		    val t3 = t2 div charSize
		    val c3 = mkChar(t3 mod charSize)
		    val t4 = t3 div charSize
		    val c4 = mkChar(t4 mod charSize)
		in
		  str(c4) ^ str(c3) ^ str(c2) ^ str(c1) ^ str(c0)
		end
	      
	    fun hashNamedCtype (namect as (Bindings.Struct(tid, _) | Bindings.Union(tid, _))) =
		(case Tidtab.find(anonTidtab, tid) of  (* anonTidtab is a memo table for struct name printing *) 
		   SOME str => str
		 | NONE =>
		     let val ctFieldNameList =
		       case namect of
			 Bindings.Struct(_, ctmemoliList) =>
			   let fun f (ct, NONE, _) = (ct, "")
				 | f (ct, SOME({name, ...}: Ast.member), _) = (ct, Symbol.name name)
			   in
			     List.map f ctmemoliList
			   end
		       | Bindings.Union(_, ctMemList) =>
			   let fun f (ct, {name, ...}:Ast.member) = (ct, Symbol.name name) in
			     List.map f ctMemList
			   end
		       | _ => nil
			   
			 fun f (hash, nil) = hash
			   | f (hash, c :: cl) = f(IntInf.mod(IntInf.+(IntInf.*(hash,IntInf.fromInt 67), IntInf.fromInt(ord c)), anonHashSize), cl)
			 (* mod should be andb - can't find IntInf.andb *)
			     
			 fun g((ct, fieldName) :: l, soFar) = 
			     let 
				 val soFar = f(soFar, explode fieldName)
				 val {bytes=size, ...} = Sizeof.byteSizeOf {err=(fn _ => ()), warn=(fn _ => ()), bug=(fn _ => ()), sizes=Sizes.defaultSizes} tidtab ct
				 val soFar = f(soFar, explode (Int.toString(size)))
			     in
			       g(l, soFar)
			     end
			   | g(nil, soFar) = soFar
			 val hash = g(ctFieldNameList, IntInf.fromInt 0)
			 val hash = ( (IntInf.toInt hash)
				     handle _ => (print("Internal Error: hash overflow\n");
						  (* should never happen, since we mod with anonHashSize *)
						  0))
			   
			 val str = (case namect of Bindings.Struct _ => "S" | Bindings.Union _ => "U" | _ => "")
			   ^ (asciFy hash)
		     in
			       Tidtab.insert(anonTidtab, tid, str);
			       str
		     end)
		
	      | hashNamedCtype (Bindings.Enum _) = "enum"  (* should never happen *)
	      | hashNamedCtype (Bindings.Typedef _) = "typedef" (* should never happen *)
		
	    fun printMember ({name, ...} : Ast.member) =
		(case Symbol.kind name of
		   Symbol.MEMBER tid =>
		     let
			 val strct =
			   case Tidtab.find (tidtab, tid)
			     of (SOME{name=SOME name, ...}) : Bindings.tidBinding option => name
			   | SOME{name=NONE, ntype=SOME nct, ...} => "anon_" ^ (hashNamedCtype nct)
			   | SOME{name=NONE, ntype=NONE, ...} => "anon1"
			   | NONE => "anon2" 
		 val name = (Symbol.name name)
		 val fullName = if strct = "" then name else (strct ^ "." ^ name)
	       in
		 fullName
	       end
	   | _ => "anon3")
	     
	fun printMember2 (mem as {name, uid, ...} : Ast.member) =
	  (* provides matching for common subsequences of structs *)
	  (* where common == same sequence of type sizes *)
	  (case Pidtab.find(memberPidtab, uid) of
	     SOME str => str
	   | NONE =>
	       (case Symbol.kind name of
		  Symbol.MEMBER tid =>
		    (case Tidtab.find (tidtab, tid) of
		       SOME{ntype=SOME(Bindings.Union _), ...} => printMember mem
		     | SOME{ntype=SOME(Bindings.Struct(tid, ctmemoliList)), ...} => 
			 let fun f ((ct, nameOpt, li) :: l) = 
			   let
			     val i = case li of
			       NONE => #bits(Sizeof.bitSizeOf{err=(fn _ => ()),
							      warn=(fn _ => ()),
							      bug=(fn _ => ()),
							      sizes=Sizes.defaultSizes} tidtab ct)
			     | SOME i => Int32.toInt i
			   in
			     case nameOpt of
			       SOME({name=name', ...}:Ast.member) =>
				 if (Symbol.name name') = (Symbol.name name) then [i] else i :: (f l)
			     | NONE => i :: (f l)
			   end
			       | f nil = nil
			     fun hasher (hash, nil) = hash
			       | hasher (hash, i :: il) = hasher(IntInf.mod(IntInf.+(IntInf.*(hash,IntInf.fromInt 67), IntInf.fromInt i), anonHashSize), il)
			 (* mod should be andb - can't find IntInf.andb *)
			 in
			   asciFy(IntInf.toInt (hasher(IntInf.fromInt 0, f ctmemoliList)))
			 end
		     | _ => "???1")
		| _ => "???2"))
	     
	fun printMember3 (mem as {name, uid, ...} : Ast.member) =
	  (* provides matching for common subsequences of structs *)
	  (* where common == same sequence of types *)
	  (case Pidtab.find(memberPidtab, uid) of
	     SOME str => str
	   | NONE =>
	       (case Symbol.kind name of
		  Symbol.MEMBER tid =>
		    (case Tidtab.find (tidtab, tid) of
		       SOME{ntype=SOME(Bindings.Union _), ...} => printMember mem
		     | SOME{ntype=SOME(Bindings.Struct(tid, ctmemoliList)), ...} => 
			 let fun f ((ct, nameOpt, li) :: l) = 
			   let
			     val ctString = ctypeToString tidtab (TypeUtil.getCoreType tidtab ct)
			     val iString =
			       case li of
				 NONE => ""
			       | SOME i => Int32.toString i
			     val s = ctString ^ iString
			   in
			     case nameOpt of
			       SOME({name=name', ...}:Ast.member) =>
				 if (Symbol.name name') = (Symbol.name name) then s else s ^ (f l)
			     | NONE => s ^ (f l)
			   end
			       | f nil = ""
			     fun hasher (hash, nil) = hash
			       | hasher (hash, c :: cl) = hasher(IntInf.mod(IntInf.+(IntInf.*(hash,IntInf.fromInt 67), IntInf.fromInt(ord c)), anonHashSize), cl)
			     (* mod should be andb - can't find IntInf.andb *)
			     val cl = String.explode(f ctmemoliList)
			 in
			   asciFy(IntInf.toInt (hasher(IntInf.fromInt 0, cl)))
			 end
		     | _ => "???1")
		| _ => "???2"))
	     
	val printMember = (case (commonSizeFields, commonTypeFields)
			     of (false, false) => printMember
			   | (true, false) => printMember2
			   | (false, true) => printMember3
			   | (true, true) => (print "Error: specified both commonSizeFields and commonTypeFields: commonTypeFields wins!\n"; printMember3))
	  
	  
	  
	  
	  
	  
	  
	  
	  
	  
	  
	  
	  
	  
	  
	  
	  
	
	(* Consts *)
      
	exception constNotFound
	
	val constsTable =
	  ConstMap.mkTable (100, constNotFound): loc list ConstMap.hash_table
	  
	fun insertConst (cst, loc) =
	  let val list =
	    ConstMap.lookup constsTable cst
	    handle constNotFound => nil
	  in
	    ConstMap.insert constsTable (cst, loc::list)
	  end
	
	
	(* Vars *)
      
	exception varNotFound
	
	val varsTable = UidMap.mkTable (100, varNotFound)
	  
	fun insertVar (v, info) =
	  let val {var, dep, rev, context, used, proto} =
	    UidMap.lookup varsTable (varToPid v)
	    handle varNotFound =>
	      {var=v, dep=nil, rev=nil, context=NONE, used=false, proto=NONE}
	  in
	    UidMap.insert varsTable
	    (varToPid v, (case info of 
			    Dep d => {dep=d::dep, rev=rev, context=context,
				      used=true, var=var, proto=proto}
			  | Rev r => {dep=dep, rev=r::rev, context=context,
				      used=true, var=var, proto=proto}
			  | Used => {dep=dep, rev=rev, context=context,
				     used=true, var=var, proto=proto}
			  | Context f => {dep=dep, rev=rev, context=SOME f,
					  used=used, var=var, proto=proto}
			  | Proto p => {dep=dep, rev=rev, context=context,
					used=true, var=var, proto=SOME p}))
	  end
	
	
	(* Static deps *)
      
	val staticDeps = ref (nil: staticDep list)
	  
	fun emitStatic (dep as Addrof {write, lval, ...}) =
	  (insertVar (write, Used);
	   insertVar (V lval, Used);
	   staticDeps := dep :: (!staticDeps))
	  
	  
	(* Deps *)
	  
	fun combineDC (NOTHING, _) = NOTHING
	  | combineDC (_, NOTHING) = NOTHING
	  | combineDC (WEAK, _) = WEAK
	  | combineDC (_, WEAK) = WEAK
	  | combineDC (STRONG, _) = STRONG
	  | combineDC (_, STRONG) = STRONG
	  | combineDC (STRONG_LVAL, STRONG_LVAL) = STRONG_LVAL
	  
	fun emitSimple {write:rval, read:rval, loc:loc} =
	  let
	    fun addDep {w, wd=false, r, rd=false, lstr:string, rstr:string, dcat, loc} = 
		let val tmp = A {pid=Pid.new(), loc=loc} in
		    addDep {w=w, wd=false, r=tmp, rd=true, lstr=lstr, rstr=rstr, dcat=dcat, loc=loc};
		    addDep {w=tmp, wd=true, r=r, rd=false, lstr=lstr, rstr=rstr, dcat=dcat, loc=loc}
		end
	    | addDep {w, wd, r, rd, lstr:string, rstr:string, dcat, loc} = 
	      let val dep =
		SimpleDep {write=w, writeDirect=wd, read=r, readDirect=rd,
			   loc=loc, dcat=dcat, lstr=lstr, rstr=rstr}
	      in
		insertVar (r, Dep dep);
		insertVar (w, Rev dep)
	      end
	  
	    fun cstr ("", lstr, rstr, "") = (lstr, rstr)
	      | cstr (lstrOld, lstrNew, rstrNew, rstrOld) =
	      (lstrOld ^ "(" ^ lstrNew, rstrNew ^ ")" ^ rstrOld)
	      
	    fun trav _ Nothing = ()
	      | trav _ (Const _) = ()
	      | trav {w, wd, lstr, rstr, direct, dcat} (Var var) =
		addDep {w=w, wd=wd, r=var, rd=direct, lstr=lstr, rstr=rstr,
			dcat=dcat, loc=loc}
	      | trav {w, wd, lstr, rstr, direct, dcat}
		(AddrOfId {vid, loc=tmp, adornment}) =
		(if wd then emitStatic (Addrof {write=w, lval=vid, loc=loc})
		 (* ignore lstr and rstr *)
		 else
		   let val temp = A {pid=Pid.new(), loc=tmp} in
		     emitStatic (Addrof {write=temp, lval=vid, loc=loc});
		     addDep {w=w, wd=wd, r=temp, rd=true, lstr="", rstr="",
			     dcat=STRONG_LVAL, loc=loc}
		   end)
	      | trav {w, wd, lstr, rstr, direct=true, dcat}
		(DeRef {lstr=lstr', rval, rstr=rstr', ...}) =
		let val (lstr, rstr) = cstr (lstr, lstr', rstr', rstr) in
		  trav {w=w, wd=wd, lstr=lstr, rstr=rstr, direct=false,
			dcat=dcat} rval
		end
	      | trav {w, wd, lstr, rstr, direct=false, dcat}
		(DeRef {lstr=lstr', rval, rstr=rstr', adornment, loc=tmp}) =
		let val temp = A {pid=Pid.new(), loc=tmp} in
		  addDep {w=w, wd=wd, r=temp, rd=false, lstr=lstr, rstr=rstr,
			  dcat=dcat, loc=loc};
		  trav {w=temp, wd=true, lstr=lstr', rstr=rstr', direct=false,
			dcat=STRONG_LVAL} rval
		end
	      | trav {w, wd, lstr, rstr, direct, dcat}
		(Op1 {lstr=lstr', rval, dcat=dcat', rstr=rstr'}) =
		let val (lstr, rstr) = cstr (lstr, lstr', rstr', rstr) in
		  trav {w=w, wd=wd, lstr=lstr, rstr=rstr, direct=direct,
			dcat=combineDC (dcat, dcat')} rval
		end
	      | trav {w, wd, lstr, rstr, direct, dcat}
		(Op2 {rval1, dcat1, str, dcat2, rval2}) =
		let
		    val (lstr1, rstr1) = cstr (lstr, "", str^"#", rstr)
		    val (lstr2, rstr2) = cstr (lstr, "#"^str, "", rstr)
	      in
		  trav {w=w, wd=wd, lstr=lstr1, rstr=rstr1, direct=direct,
		      dcat=combineDC (dcat, dcat1)} rval1;
		  trav {w=w, wd=wd, lstr=lstr2, rstr=rstr2, direct=direct,
			dcat=combineDC (dcat, dcat2)} rval2
		end
	      | trav {w, wd, lstr=lstr, rstr=rstr, direct, dcat}
		(Join {rval1, rval2}) =
		(trav {w=w, wd=wd, lstr=lstr, rstr=rstr, direct=direct,
		       dcat=dcat} rval1;
		 trav {w=w, wd=wd, lstr=lstr, rstr=rstr, direct=direct,
		       dcat=dcat} rval2)
		

(* Note: we collect dcats from operations inside *;
	   if *(...) depends on x, then x must be STRONG_LVALS anyway *)



	  in
		  case (write, read) of
		    (Nothing, _) => ()
		  | (_, Nothing) => ()
		  | (_, Const _) => ()  (* future feature: utilize constants? *)
		
	    | (Var var, _) =>
		trav {w=var, wd=true, lstr="", rstr="",
		      direct=true, dcat=STRONG_LVAL} read
	    | (DeRef {rval=Var var, ...}, _) =>
		trav {w=var, wd=false, lstr="", rstr="",
		      direct=true, dcat=STRONG_LVAL} read
	    | (DeRef {rval, adornment, loc=tmp, ...}, _) => 
		let val temp = A {pid=Pid.new(), loc=tmp} in
		  trav {w=temp, wd=false, lstr="", rstr="",
			direct=true, dcat=STRONG_LVAL} read;
		  trav {w=temp, wd=true, lstr="", rstr="",
			direct=true, dcat=STRONG_LVAL} rval
		end
	      
	    (* There is a bug here somewhere
	     Join error is generated on series2 codebase?? *)
	    | (Op1 _, _) => print "Internal Error: Op1 case in lhs."
	    | (Op2 _, _) => print "Internal Error: Op2 case in lhs."
	    | (Const _, _) => print "Internal Error: Const case in lhs."
	    | (Join _, _) => print "Internal Error: Join case in lhs."
	    | (AddrOfId _, _) => print "Internal Error: AddrOfId case in lhs."
	  end
	
	
	(* Functions *)
      
	val inFunction = ref (NONE: vid option) (* Enclosing function *)
	val returns = ref (nil: (rval * loc) list)      (* rval returned *)
	  
	fun buildRet _ ret false = ret
	  | buildRet var NONE _ =
	    let val r = R {pid=Pid.new(), var=var} in
	      insertVar (r, Used);
	      SOME r
	    end
	  | buildRet _ ret _ = ret
	
	fun buildArgs var =
	  let
	    fun build _ [] 0 = []
	      | build n (x::y) 0 = x::(build (n+1) y 0)
	      | build n (x::y) p = x::(build (n+1) y (p-1))
	      | build n [] p =
	      let val f = F {pid=Pid.new(), arg=n, var=var} in
		insertVar (f, Used);
		f::(build (n+1) [] (p-1))
	      end
	  in
	    build 1
	  end
	  
	fun updateProto var {ptr: bool, ret: bool, args: int} =
	  let
	    val (r, a) =
	      case UidMap.find varsTable (varToPid var) of
		SOME {proto=SOME {ret, args, ...}, ...} => (ret, args)
	      | _ => (NONE, [])
	    val r = buildRet var r ret
	    val a = buildArgs var a args
	    val proto = {ptr=ptr, ret=r, args=a}
	  in
	    insertVar(var, Proto proto);
	    proto
	  end
	
	fun emitFunDef vid rvals params loc =
	  let
	    val {ret, args, ...} =
	      updateProto (V vid) {ptr=false, ret=not (null rvals),
			       args=List.length params}
	    fun emitRet (SOME r) (rval, rloc) =
	      emitSimple {write=Var r, read=rval, loc=rloc}
	      | emitRet _ _ = ()
	    fun emitArg (arg, prm) =
	      emitSimple {write=Var (V prm),
			  read=Var arg, loc=loc}
	  in
	    insertVar (V vid, Proto {ptr=false, ret=ret, args=args});
	    List.app (emitRet ret) rvals;
	    ListPair.app emitArg (args, params)
	  end
	
	
	(* Vids *)
	  
	fun idToVid ({name, uid, location, ctype, global, ...}: Ast.id): vid =
	  {pid=uid, name=Symbol.name name, tp=ctype, loc=location,
	   global=global}
	  
	fun memberToVid (member as
			 {name, uid, location, ctype, ...}: Ast.member): vid =
	  let
	    val str = printMember member
	    val global = (case Symbol.kind name
			    of Symbol.MEMBER tid =>
			      (case Tidtab.find (tidtab, tid)
				 of SOME {global, ...} => global
			       | NONE =>
				   (print "Internal error: can't find tid\n";
				    true))
			  | _ =>
			      (print "Internal error: tid is not a member\n";
			       true))
	  in
	    {pid=uid, name=str, tp=ctype, loc=location, global=global}
	  end
	
	
	(* Types *)
      
	fun getType adornment = 
	  (case Aidtab.find(aidtab, adornment) of
	     SOME ty => ty
	   | NONE => (print "Internal Error: no type for this expression.";
		      print (Int.toString adornment);
		      print "\n";
		      Ast.Void))
	  
	fun isArray ({tp, ...}: vid) = TypeUtil.isArray tidtab tp
	  
	fun isFunction ({tp, ...}: vid) = 
	  TypeUtil.isNonPointerFunction tidtab tp
	  
	fun checkArray (loc, adornment) vid =
	  if isArray vid then AddrOfId {vid=vid, loc=loc, adornment=adornment}
	  else Var (V vid)
	    
	fun checkFunctionOrArray (loc, adornment) vid =
	  if isFunction vid orelse isArray vid then
	    AddrOfId {vid=vid, loc=loc, adornment=adornment}
	  else Var (V vid)
	    
	    
	(* Main *)
	
	fun deref tidtab adornment = (case (TypeUtil.deref tidtab (getType adornment)) of SOME x => x | NONE => Ast.Void)

	fun derefRval (adornment, _, _, _) (rval as AddrOfId {vid, ...}) =
	    if TypeUtil.isFunction tidtab (deref tidtab adornment) then rval else Var (V vid)
	  | derefRval (adornment, loc, lstr, rstr) rval =
	    if TypeUtil.isFunction tidtab (deref tidtab adornment) then rval else
		DeRef {lstr=lstr, rval=rval, rstr=rstr,
		       adornment=adornment, loc=loc}
	  
	fun travDecl (TypeDecl _) = ()
	  | travDecl (VarDecl (id, initExpr)) =
	  let val vid as {global, ...} = idToVid id in
	    (case initExpr of 
	       SOME (Simple (expr as EXPR (_, _, loc))) =>	      
		 emitSimple {write=Var (V vid),
			     read=travExpr (expr, {throwAway=false}),
			     loc=loc}
	     | _ => ());
	    (case !inFunction of
	       SOME function =>
		 if not global then (insertVar(V vid, Context function);
				     insertVar(V function, Used))
		 else ()
	     | NONE => ())
	  end
	
	and travEdecl (DECL (ExternalDecl decl, _, _)) = travDecl decl
	  | travEdecl (DECL (ExternalDeclExt _, _, _)) = ()
	  | travEdecl (DECL (FunctionDef (id, idlist, statement), _, loc)) =
	  let val vid = idToVid id in
	    inFunction := SOME vid;
	    returns := nil;
	    travStat statement;
	    inFunction := NONE;
	    emitFunDef vid (!returns) (List.map idToVid idlist) loc;
	    returns := nil
	  end

	and emitCallSite calle loc =
	    (case (!inFunction) of
		 SOME vid =>
		     (let val dep = CallSite{caller=vid, calle=calle, loc=loc} in
			  insertVar(V vid, Dep dep);
			  insertVar(calle, Rev dep)
		      end)
	       | NONE => ())
	
	and travStat (STMT(coreStat, _, loc)) =
	  case coreStat of
	    Expr NONE => ()
	  | Expr (SOME (EXPR (Call (EXPR (Id id, _, _), exprl), _, loc))) =>
	      let 
		val vid as {pid, tp, ...} = idToVid id
		fun emitArg (arg, expr) =
		  let val rval = travExpr (expr, {throwAway=false}) in
		    emitSimple {write=Var arg, read=rval, loc=loc}
		  end
		val ptr = not (TypeUtil.isNonPointerFunction tidtab tp)
		val {args, ...} =
		  updateProto (V vid) {ptr=ptr, ret=false,
				       args=List.length exprl}
	      in
		emitCallSite (V vid) loc;
		ListPair.app emitArg (args, exprl)
	      end
	  | Expr (SOME expr) => (travExpr (expr, {throwAway=true}); ())
	  | Compound (decL, statL) =>
	      (List.map travDecl decL; List.map travStat statL; ())
	  | While (expr, stat) => 
	      (travExpr (expr, {throwAway=true}); travStat stat)
	  | Do (expr, stat) =>
	      (travExpr (expr, {throwAway=true}); travStat stat)
	  | For (expr1, expr2, expr3, stat) => 
	      (travExprOpt(expr1, {throwAway=true});
	       travExprOpt(expr2, {throwAway=true});
	       travExprOpt(expr3, {throwAway=true});
	       travStat stat)
	  | Labeled (pid, stat) => travStat stat
	  | CaseLabel (lint, stat) => travStat stat
	  | DefaultLabel stat => travStat stat
	  | Goto pid => ()
	  | Break => ()
	  | Continue => () 
	  | Return expr =>
	      (case (travExprOpt (expr, {throwAway=false}), !inFunction) of
		 (Nothing, _) => ()
	       | (v, SOME _) => returns := (v, loc) :: (!returns)
	       | (v, NONE) =>
		   warning "Warning: function return not in function?\n")
	  | IfThen (expr, stat) =>
	      (travExpr (expr, {throwAway=true}); travStat stat)
	  | IfThenElse (expr, stat1, stat2) => 
	      (travExpr (expr, {throwAway=true});
	       travStat stat1;
	       travStat stat2)
	  | Switch (expr, stat) => 
	      (travExpr (expr, {throwAway=true}); travStat stat)
	  | StatExt _ => ()
	  | ErrorStmt => ()
	      
	and travExprOpt (SOME expr, params) = travExpr (expr, params)
	  | travExprOpt (NONE, _) = Nothing
	  
	and travExpr (EXPR (core_expr, adornment, loc), ta as {throwAway}) =
	  case core_expr of
	    IntConst lint => (insertConst (Int32.toString lint, loc);
			      Const(Int32.toString lint))
	  | RealConst real => Const(Real.toString real)
	  | StringConst string => 
		let  (* allocate new memory region for this string *)
		    val uid = Pid.new()
		    val tp = Ast.Pointer(Ast.Void)
		    val name = "string"
		    val loc = loc
		    val global = false
		    val vid = {pid=uid, name=name, tp=tp, loc=loc, global=global}
		in
		    insertConst (string, loc);
		    if noStrings then Const string else AddrOfId{vid=vid, loc=loc, adornment=adornment}
		end
	  | Cast (ctype, expr) => travExpr (expr, ta)
	  | Id id => checkFunctionOrArray (loc, adornment) (idToVid id)
	  | EnumId (id, lint) => Const (Int32.toString lint)
	  | SizeOf ct => Const ("sizeof(" ^ (ctypeToString tidtab ct) ^ ")")
	  | ExprExt _ => Nothing
	  | ErrorExpr => Nothing
	  | Call (EXPR (Id id, _, _), exprl) => (* direct calls *)
	      let 
		val vid as {name, pid, tp, ...} = idToVid id
		fun memberStr (x: string, y::l) =
		  (x=y) orelse memberStr (x, l)
		  | memberStr (x, nil) = false
	      in
		if checkForMallocs andalso memberStr(name, mallocs) then
		  let (* allocate new memory region for this malloc call *)
		    val vid = {pid=Pid.new(), name=name, global=false,
			       tp=Ast.Pointer(Ast.Void), loc=loc}
		  in
		    AddrOfId {vid=vid, loc=loc, adornment=adornment}
		  end
		else
		  let
		    fun emitArg (arg, expr) =
		      let val rval = travExpr (expr, {throwAway=false}) in
			emitSimple {write=Var arg, read=rval, loc=loc}
		      end
		    val ptr = not (TypeUtil.isNonPointerFunction tidtab tp)
		    val {ret, args, ...} =
		      updateProto (V vid) {ptr=ptr, ret=not throwAway,
					   args=List.length exprl}
		  in
		    ListPair.app emitArg (args, exprl);
		    emitCallSite (V vid) loc;
		    case ret of
		      SOME r => Var r
		    | NONE => Nothing
		  end
	      end
	  | Call (expr, exprl) =>
	      let
		val temp = A {pid=Pid.new(), loc=loc}
		val rval = travExpr (expr, {throwAway=false})
		fun emitArg (arg, expr) =
		  let val rval = travExpr (expr, {throwAway=false}) in
		    emitSimple {write=Var arg, read=rval, loc=loc}
		  end
		val {ret, args, ...} =
		  updateProto temp {ptr=true, ret=not throwAway,
				    args=List.length exprl};
	      in
		emitSimple {write=Var temp, read=rval, loc=loc};
		ListPair.app emitArg (args, exprl);
		emitCallSite temp loc;
		case ret of
		  SOME r => Var r
		| NONE => Nothing
	      end
	    
	  | QuestionColon (expr1, expr2, expr3) =>
	      let
		val _ = travExpr (expr1, {throwAway=true})
		val v2 = travExpr (expr2, ta)
		val v3 = travExpr (expr3, ta)
	      in
		Join {rval1=v2, rval2=v3}
	      end
	    
	  | Assign (expr1, expr2) => 
	      let
		val v1 = travExpr (expr1, {throwAway=false})
		val v2 = travExpr (expr2, {throwAway=false})
	      in
		emitSimple {write=v1, read=v2, loc=loc};
		v1
	      end
	    
	  | Comma (expr1, expr2) =>
	      let
		val v1 = travExpr (expr1, {throwAway=true})
		val v2 = travExpr (expr2, ta)
	      in
		v2
	      end
	    
	  | Sub(expr1, expr2) =>
	      (* treat arrays in index independent manner *)
	      let
		val v1 = travExpr (expr1, ta)
		val _ = travExpr (expr2, {throwAway=true})
		val arrayTp = TypeUtil.isArray tidtab (getType adornment)
	      in
		if arrayTp then v1
		(* partial array application -- no deref *)
		else
		  (* full array application -- add deref *)
		  derefRval (adornment, loc, "", "[]") v1
	      end
	    
	  | Member (expr, member) =>
	      (if noFields then 
		 travExpr (expr, ta)
	       else
		 let
		   val v = travExpr (expr, {throwAway=true})
		   val vid: vid = memberToVid member
		 in
		   checkArray (loc, adornment) vid
		 end)
	      
	  | Arrow (expr, member) =>
	      (if noFields then
		 derefRval (adornment, loc, "*", "") (travExpr (expr, ta))
	       else 
		 let
		   val v = travExpr (expr, {throwAway=true})
		   val vid = memberToVid member
		 in
		   checkArray (loc, adornment) vid
		 end)
	      
	  | Deref expr => 
	      derefRval (adornment, loc, "*", "") (travExpr (expr, ta))
	      
	  | AddrOf expr =>
	      (case travExpr (expr, ta) of
		 Var (V vid) =>
		   AddrOfId {vid=vid, loc=loc, adornment=adornment}
	       | DeRef {rval, ...} => rval 
	       | rval as (AddrOfId {vid, ...}) =>
		   (if isArray vid orelse isFunction vid then rval
		    else (warning "Warning: addrof problem (1)\n"; Nothing))
	       | Nothing => (warning "Warning: addrof problem: Nothing\n";
			     Nothing)
	       | Op1 _ => (warning "Warning: addrof problem Op1\n";
			   Nothing)
	       | Op2 _ => (warning "Warning: addrof problem Op2\n";
			   Nothing)
	       | Join _ => (warning "Warning: addrof problem Join\n";
			    Nothing)
	       | Const _ => (warning "Warning: addrof problem Const\n";
			     Nothing)
	       | Var _ => (warning "Warning: addrof problem: Var\n";
			   Nothing))
	      
	  | Binop(oper,expr1, expr2) =>
	      let 
		val x = case oper of
		  Plus =>  SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str="+", assign=false, sym=true}
		| Minus => SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG,      str="-", assign=false, sym=true}
		| Times => SOME{Ldcat=WEAK,        Rdcat=WEAK,        str="*", assign=false, sym=true}
		| Divide =>SOME{Ldcat=WEAK,        Rdcat=NOTHING,     str="/", assign=false, sym=false}
		| Mod =>   SOME{Ldcat=WEAK,        Rdcat=NOTHING,     str="%", assign=false, sym=false}
		| Gt =>    NONE
		| Lt =>    NONE
		| Gte =>   NONE
		| Lte =>   NONE
		| Eq =>    NONE
		| Neq =>   NONE
		| And =>   NONE
		| Or =>    NONE
		| BitOr => SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str= "|", assign=false, sym=true}
		| BitAnd =>SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str= "&", assign=false, sym=true}
		| BitXor =>SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str= "^", assign=false, sym=true}
		| Lshift =>SOME{Ldcat=WEAK,        Rdcat=NOTHING,     str= ">>", assign=false, sym=false}
		| Rshift =>SOME{Ldcat=WEAK,        Rdcat=NOTHING,     str= "<<", assign=false, sym=false}
			   
		| PlusAssign =>  SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str="", assign=true, sym=true}
		| MinusAssign => SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG,      str="", assign=true, sym=true}
		| TimesAssign => SOME{Ldcat=WEAK,        Rdcat=WEAK,        str="", assign=true, sym=true}
		| DivAssign =>   SOME{Ldcat=WEAK,        Rdcat=NOTHING,     str="", assign=true, sym=false}
		| ModAssign =>   SOME{Ldcat=WEAK,        Rdcat=NOTHING,     str="", assign=true, sym=false}
		| XorAssign =>   SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str="", assign=true, sym=true}
		| OrAssign =>    SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str="", assign=true, sym=true}
		| AndAssign =>   SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str="", assign=true, sym=true}
		| LshiftAssign =>SOME{Ldcat=WEAK,        Rdcat=NOTHING,     str="", assign=true, sym=false}
		| RshiftAssign =>SOME{Ldcat=WEAK,        Rdcat=NOTHING,     str="", assign=true, sym=false}
		| BinopExt _ => NONE
		val asn = case x of SOME{assign, ...} => assign | NONE => false
		val v1 = travExpr(expr1, if asn then {throwAway=false} else ta)
		val v2 = travExpr(expr2, if asn then {throwAway=false} else ta)
		val x = if alwaysTrackLvals
			  then
			    SOME{Ldcat=STRONG_LVAL, Rdcat=STRONG_LVAL, str="", assign=false, sym=false}
			else x
	      in case x of
		NONE => Nothing
	      | SOME{Ldcat, Rdcat, str, assign, sym} =>
		  let val opResult = 
		    case (v1, v2, Rdcat) of
		      (Const str1, Const str2, _) => Const("(" ^ str1 ^ str ^ str2 ^ ")")
		    | (Nothing, Nothing, _) => Nothing
		    | (Nothing, _, NOTHING) => Nothing
		    | (Nothing, something, _) => Op1{lstr="#"^str, rval=v2, rstr="", dcat=Rdcat}  (* Nothing case? *)
		    | (Const _, _, NOTHING) => Nothing
		    | (Const str1, something, _) => Op1{lstr=str1^str, rval=v2, rstr="", dcat=Rdcat}
		    | (something, _, NOTHING) => Op1{lstr="", rval=v1, rstr=str^"#", dcat=Ldcat}  (* Nothing case? *)
		    | (something, Nothing, _) => Op1{lstr="", rval=v1, rstr=str^"#", dcat=Ldcat}  (* Nothing case? *)
		    | (something, Const str2, _) => Op1{lstr="", rval=v1, rstr=str^str2, dcat=Ldcat}
		    | (_, _, _) => Op2{rval1=v1, dcat1=Ldcat, str=str, rval2=v2, dcat2=Rdcat}
		  in
		    if assign then 
		      (emitSimple{write=v1, read=opResult, loc=loc};
		       v1)
		    else
		      opResult
		  end
	      end
		   
	  | Unop(unop, expr) =>
	      let 
		val x = case unop of
		  Uplus =>  SOME{dcat=STRONG_LVAL, lstr="+", rstr="", assign=false}
		| Negate => SOME{dcat=STRONG, lstr="-", rstr="", assign=false}
		| Not =>    NONE
		| BitNot => SOME{dcat=STRONG,        lstr="~",  rstr="",     assign=false}
		| PreInc => SOME{dcat=STRONG_LVAL, lstr="++", rstr="",     assign=true}
		| PostInc =>SOME{dcat=STRONG_LVAL, lstr="",      rstr="++",assign=true}
		| PreDec => SOME{dcat=STRONG_LVAL, lstr="--", rstr="",     assign=true}
		| PostDec =>SOME{dcat=STRONG_LVAL, lstr="",      rstr="--",assign=true}
		| UnopExt _ => NONE
		val asn = case x of SOME{assign, ...} => assign | NONE => false
		val v = travExpr(expr, ta)
	      (* only assign ops are ++, --,
	       so ok if we ignore these wrt dep analysis*)
	      in case x of
		NONE => Nothing
	      | SOME{dcat, lstr, rstr, assign} =>
		  let val opResult =
		    case v of
		      Nothing => Nothing 
		    | Const str1 => Const("("^lstr^str1^rstr^")")
		    | _ => Op1{lstr=lstr, rval=v, rstr=rstr, dcat=dcat}
		  in
		    opResult
		  end
	      end
      in
	List.app travEdecl edl;
	(UidMap.listItems varsTable, !staticDeps,
	 ConstMap.listItemsi constsTable)
      end (* travAst *)
  end (* struct *)
