/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <strings.h>
#include "fatal.h"
#include "bigarray.h"

#define BITS 12
#define SIZE (1 << BITS)
#define MASK (SIZE - 1)
#define INCR (1 << 6)

struct bigarray {
  int nbObjects;         /* nb of objects */
  int objectSize;        /* size of objects */
  int space;
  unsigned char **array;
 /* size = nbObjects * objectSize bytes */
};

struct bigarray *createArray(int objectSize) {
  struct bigarray *array;

  array = (struct bigarray *)MALLOC(sizeof(struct bigarray));
  array->nbObjects = 0;
  array->objectSize = objectSize;
  array->space = 0;
  array->array = 0;
  return array;
}

void *newElement(struct bigarray *array) {
  int quot, rem;

  assert(array);

  quot = array->nbObjects >> BITS;
  rem = array->nbObjects & MASK;
  
  if(!rem) {
    if(quot == array->space) {
      array->space = quot + INCR;
      array->array = (unsigned char **)REALLOC(array->array, array->space * sizeof(unsigned char *));
    }
    array->array[quot] = (unsigned char *)CALLOC(SIZE, array->objectSize);
  }
  
  array->nbObjects++;

  return array->array[quot] + (array->objectSize * rem);
}

void *getElement(struct bigarray *array, int index) {
  assert(array);

  if(index >= array->nbObjects) return 0;
  return array->array[index >> BITS] + (array->objectSize * (index & MASK));
}

int arraySize(struct bigarray *array) {
  assert(array);

  return array->nbObjects;
}

int getIndex(struct bigarray *array, void *element) {
  int quot, rem;
  
  assert(array);
  
  for(quot=0; quot<=(array->nbObjects-1)>>BITS; quot++) {
    rem = ((unsigned char *)element - array->array[quot]) / array->objectSize;
    if(rem < SIZE) return quot * SIZE + rem;
  }
  return -1;
}

void disposeElements(struct bigarray *array, int index) {
  int quot, rem;
  
  assert(array);

  if(index >= array->nbObjects) return;
  for(quot=((index-1)>>BITS)+1; quot<=(array->nbObjects-1)>>BITS; quot++) {
    free(array->array[quot]);
  }

  array->nbObjects = index;

  quot = index >> BITS;
  rem = index & MASK;
  if(rem) bzero(array->array[quot] + (array->objectSize * rem), array->objectSize * (SIZE - rem));
}

void freeArray(struct bigarray *array) {
  int quot;

  assert(array);

  for(quot=0; quot<=(array->nbObjects-1)>>BITS; quot++) free(array->array[quot]);
  free(array->array);
  free(array);
}
