/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "list.h"
#include "storage.h"
#include "set.h"
#include "database.h"
#include "targets.h"
#include "print.h"

int showAnonymousVars = 0;
int showLocations = 1;
int showTypes = 1;
int showStrings = 1;

#define programVar(infos) (*(infos)->name && *(infos)->ctype)

static FILE *stream;

static void sorted(int unique) {
  FILE *p;

  if(unique) p = popen("sort -u", "w"); else p = popen("sort", "w");
  if(!p) fprintf(stderr, "Error: unable to launch sort\n"); else stream = p;
}

static void unsorted(void) {
  if (stream != stdout) pclose(stream);
  stream = stdout;
}

static void printLoc(struct location *loc) {
  if(loc->startLine == loc->endLine) {
    fprintf(stream, "<%s:%d>", loc->file, loc->startLine);
  } else {
    fprintf(stream, "<%s:%d-%d>", loc->file, loc->startLine, loc->endLine);
  }
}

static void printType(char* type) {
  int i;
  
  if(*type) {
    for(i=0; type[i] || type[i] == ' '; i++);
    if(type[i]) {
      fprintf(stream, "/(%s)", type); /* string has blanks: surround by parens */
    } else {
      fprintf(stream, "/%s", type); /* simple type - omit parens */
    }
  }
}

static void printVarInfos(struct varInfos *infos) {
  fprintf(stream, "%s", *infos->name?infos->name:"_tmp");
  if(showTypes) printType(infos->ctype);
  if(showLocations) {
    fprintf(stream, " ");
    printLoc(&infos->loc);
  }
}

static void printStrings(char *name, char *lstr, char*rstr) {
  if(lstr && (*lstr || *rstr)) fprintf(stream, "{%s%s%s} ", lstr, *name?name:"_tmp", rstr);
}

static void printLvals(struct variable *var) {
  struct varInfos infos[1];
  struct list *lvals;
  struct variable *lval;
  
  fprintf(stream, ":\n");
  sorted(1);
  for(lvals=getElems(var->lvals); lvals; lvals=lvals->next) {
    lval = (struct variable*)lvals->elem;
    getVarInfos(lval->offset, infos);
    if(showAnonymousVars || programVar(infos)) {
      fprintf(stream, "    ");
      printVarInfos(infos);
      fprintf(stream, "\n");
    }
  }
  unsorted();
}

static void printRevDeps(struct variable *var) {
  struct varInfos infos[1];
  struct chain *chain;
  int i;
  
  fprintf(stream, ":\n");
  sorted(1);
  for(i=0; i<chainCount(); i++) {
    chain = chainOfIndex(i);
    if(chain->readVar == var) {
      getVarInfos(chain->writeVar->offset, infos);
      if(showAnonymousVars || programVar(infos)) {
	fprintf(stream, "    ");
	printVarInfos(infos);
	fprintf(stream, "\n");
      }
    }
  }
  unsorted();
}

static int hasRevDeps(struct variable *var) {
  struct varInfos infos[1];
  struct chain *chain;
  int i;
  
  for(i=0; i<chainCount(); i++) {
    chain = chainOfIndex(i);
    if(chain->readVar == var) {
      getVarInfos(chain->writeVar->offset, infos);
      if(showAnonymousVars || programVar(infos)) return 1;
    }
  }
  return 0;
}

static int hasLvals(struct variable *var) {
  struct varInfos infos[1];
  struct list *lvals;
  struct variable *lval;
  
  if(hasNoLvals(var)) return 0;
  
  for(lvals=getElems(var->lvals); lvals; lvals=lvals->next) {
    lval = (struct variable*)lvals->elem;
    getVarInfos(lval->offset, infos);
    if(showAnonymousVars || programVar(infos)) return 1;
  }
  return 0;
}

static struct entry nextPattern;

static void printChain(struct chain *chain, int first) {
  struct varInfos infos[1];
  struct depInfos inf[1];
  
  if(!chain->offset) {
    fprintf(stream, " is a target\n");
    return;
  }

  if(isWeak(chain->dtype)) {
    fprintf(stream, " W");
    if(isViaPointer(chain->dtype)) fprintf(stream, "*");
  } else {
    if(isViaPointer(chain->dtype)) fprintf(stream, " *");
  }
  
  while(chain->offset && chain->readVar) {
    getDepInfos(chain->offset, inf);
    getVarInfos(chain->readVar->offset, infos);
    if(showAnonymousVars || programVar(infos)) {
      fprintf(stream, " -%s%s-> ", isViaPointer(inf->dtype)?"*":"", isWeak(inf->dtype)?"W":"");
      nextPattern = *(struct entry *)infos;
      infos->loc = inf->loc;
      if(showStrings) printStrings(infos->name, inf->lstr, inf->rstr);
      printVarInfos(infos);
      if(first) {
	fprintf(stream, "\n");
	return;
      }
    }
    if(isReadPtr(inf->dtype)) {
      chain = chain->readVar->lvalsChain;
    } else {
      chain = chain->readVar->chain; 
    }
  }
  if(showLocations) {
    fprintf(stream, " where ");
    getVarInfos(chain->writeVar->offset, infos);
    printVarInfos(infos);
  }
  fprintf(stream, "\n");
  return;
}

static int test(struct variable *var, int mode) {
  switch(mode) {
  case REVDEPS: return hasRevDeps(var);
  case ALLVARS: return 1;
  case DEPVARS:
  case DEPS:
  case CHAINS: return hasDep(var);
  case TARGETVARS: return isTargetVar(var);
  case NONTARGETVARS: return isNonTargetVar(var);
  case LVALS: return hasLvals(var);
  case LVALSDEPS: return hasLvalsDep(var);
  }
  return 0;
}

void printLvalsDeps(struct variable *var) {
  struct varInfos infos[1];

  getVarInfos(var->lvalsChain->writeVar->offset, infos);
  fprintf(stream, " points to ");
  printVarInfos(infos);
  fprintf(stream, "\n");
}

static void printLine(struct variable *var, struct varInfos *infos, int mode) {
  printVarInfos(infos);
  switch(mode) {
  case LVALS: printLvals(var); break;
  case CHAINS: printChain(var->chain, 0); break;
  case DEPS: printChain(var->chain, 1); break;
  case REVDEPS: printRevDeps(var); break;
  case LVALSDEPS: printLvalsDeps(var); break;
  default: fprintf(stream, "\n");
  }  
}

struct entry *printVars(int mode, int *index, int max, struct entry *pattern) {
  struct varInfos infos[1];
  struct variable *var;
  int i, count = 0;
  
  stream = stdout; /* hack */
  if(mode == ALLVARS || mode == DEPVARS || mode == TARGETVARS || mode == NONTARGETVARS) sorted(1);
  for(i=*index; i<varCount() && (!max || count < max); i++) {
    var = varOfIndex(i);
    if(test(var, mode)) {
      getVarInfos(var->offset, infos);
      if(match(pattern, (struct entry *)infos)) {
	if(showAnonymousVars || programVar(infos)) {
	  printLine(var, infos, mode);
	  count++;
	}
      }
    }
  }
  unsorted();
  if(i < varCount() && max && count >= max) {
    fprintf(stream, "...\n");
  }
  *index = i;
  return (mode == DEPS && count == 1)?&nextPattern:0;
}

static int isProgramVar(struct variable *var) {
  struct varInfos infos[1];
  
  getVarInfos(var->offset, infos);
  return programVar(infos);
}

int lvalCount(void) {
  struct variable *var;
  int i, count = 0, c = 0, size;

  for(i=0; i<varCount(); i++) {
    var = varOfIndex(i);
    if((showAnonymousVars || isProgramVar(var)) && !hasNoLvals(var)) {
      size = listSize(getElems(var->lvals));
      count += size;
      if(size) c++;
    }
  }
  if(debug) printf("programVar with lvals: %d\n", c);
  return count;
}

int depVarCount(void) {
  struct variable *var;
  int i, count = 0;
  
  for(i=0; i<varCount(); i++) {
    var = varOfIndex(i);
    if(isProgramVar(var) && hasDep(var)) count++;
  }
  return count;
}
