/**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************/

/*******  
  DESIGN NOTES ON SUB-TRANSITIVE SET IMPLEMENTATION

  1. Constraint Processing:
       Membership: just add element to baseElements list.
       Inclusion: just add to sets list;

  2. GetElem: 
       Perform depth-first traversal, caching results, and
       adding cycle links (skip).

******/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "fatal.h"
#include "list.h"
#include "storage.h"
#include "set.h"

#define DESKIP(x) for(;x->skip;x=x->skip)

int currentIterationIndex = 0;

void prL(struct list *l);
void stats(void);
int listSize(struct list *l);

/* internal only */
int memberEL(void *elem, struct list **pl, int update);
int subsetL(struct list *l, struct list **set, int update);
int eqList (struct list *l1, struct list *l2);
struct set *hashSet(struct set *s, int update);
void freeL(struct list *l);
struct list *getLvals(struct set *s, struct list *path);
void dumpSet(struct set *s);
int insertSetClean(struct set *from, struct set *to);
int insertSetStd(struct set *from, struct set *to);
void recomputeElemElems(struct list *lvals, struct set *lvalLvals);

#define CHUNK_N 12
#define CHUNK_SIZE (1<<CHUNK_N)
#define chunkIndex(i) (i>>CHUNK_N)
#define chunkOffset(i) (i & (CHUNK_SIZE-1))
#define MAX_CHUNKS 5000
struct set *sets[MAX_CHUNKS];

int nextSet = 0;

struct set *emptySet() {
  struct set *new;
  int chunk = chunkIndex(nextSet);
  int offset = chunkOffset(nextSet);

  if(!offset) {  /* need to allocate new chunk */
    if(chunk >= MAX_CHUNKS) {
      printf("Ran out of sets -- increase MAX_CHUNKS (currently %d)\n", MAX_CHUNKS);
      fatal("Out of sets\n");
    }
    sets[chunk] = (struct set *)malloc(CHUNK_SIZE * sizeof(struct set));
    if (!sets[chunk]) {
      printf("malloc failure in allocation of new set (current set count = %d\n", nextSet);
      fatal("malloc failure\n");
    }
    bzero(sets[chunk], CHUNK_SIZE * sizeof(struct set));
  }

  new = sets[chunk]+offset;
  nextSet++;
  
/*** bzero means that the following are not needed anymore
  new->baseElements = 0;
  new->sets = 0;
  new->size=0;
  new->lvals = 0;
  new->skip = 0;
  new->lastRecomputed = 0;
  new->onPath=0;
  new->lvalLvals=0;
***/
  return(new);
}

int recoverSetIndex(struct set *s) { /* maps pointer to a set back into a set index */
  int i;
  for(i=0; i<=chunkIndex(nextSet); i++) {
    if(s >= sets[i] && s < sets[i]+CHUNK_SIZE) {
      int n = i*CHUNK_SIZE + (s-sets[i]);
      if(s == sets[chunkIndex(n)] + chunkOffset(n)) {  /* double check */
	return(n);
      } else {
	printf("recoverSetIndex error 1\n");
      }
    }
  }
  printf("recoverSetIndex error 2\n");
  return(0);
}

void prSetL(struct list *l) {
  for(;l;l=l->next) {
    printf("%d ", recoverSetIndex((struct set*)(l->elem)));
  }  
  printf("\n");
}

int isSortedL(struct list *l) {
  void *elem;
  if(!l) return(1);

  elem=l->elem;

  for(l=l->next; l; l=l->next) {
    if (elem >= l->elem) {
      return(0);
    }
    elem=l->elem;
  }
  return(1);
}

void prL(struct list *l) {
  for(;l;l=l->next) {
    printf("%x ", l->elem);
  }  
  printf("\n");
}

void freeL(struct list *l) {
  struct list *l1;
  debugdec(int i = 0);
  for(;l;) {
    debug(i++);
    debug(if(i>100000) printf("loop: %d\n", l));
    debug(l->elem=(void *)0xabcdabcd);
    l1 = l;
    l = l->next;
    freeCons(l1);
  }  
}

void freeLL(struct list *l) {
  struct list *l1;
  debugdec(int i = 0);
  for(;l;) {
    debug(i++);
    debug(if(i>100000) printf("loop: %d\n", l));
    freeL(l->elem);
    debug(l->elem=(void *)0xabcdabcd);
    l1 = l;
    l = l->next;
    freeCons(l1);
  }  
}

int memberEL(void *elem, struct list **pl, int update) {
  /* Check to see if elem is in the list *pl; add elem if update set */
  struct list *l = *pl;
  debugdec(struct list **savePl = pl);

  assert(isSortedL(*pl));

  for(; l; ) {
    if(elem > l->elem) {
      pl=&(l->next);
      l=*pl;
      continue; /* not there yet */
    } else if(elem < l->elem) {
      if(update) {
        (*pl)=cons(elem, l);
      }
      assert(isSortedL(*savePl));
      return(0);
    } else /* found it. */ {
      return(1);
    }
  }

  if(update) {
    (*pl)=cons(elem, 0);
  }
  assert(isSortedL(*savePl));
  return(0);
}

int subsetL(struct list *l, struct list **set, int update) {
   /* add elements in l to s; l never changed; return count of changes */

  struct list *s = *set;
  int changes = 0;
  debugdec(struct list **saveSet = set);

  assert(isSortedL(l) && isSortedL(*set));

  for(; l && s; ) {
    if(l->elem > s->elem ) {
      set=&(s->next);
      s=*set;
      continue; 
    } else if(l->elem < s->elem ) {
      changes++;
      if(update) {
	struct list *new = cons(l->elem, s);
        (*set)=new;
	set=&(new->next);
	/* no need to change s */
	assert(s==*set);
      }
      l=l->next;
      continue;
    }
    set=&(s->next);
    s=*set;
    l=l->next;
  }
  
  /* copy extra elements */
  for(; l; l=l->next) {
    struct list *new=cons(l->elem, 0);
    (*set)=new;
    set=&(new->next);
    changes++;
  }
  assert(isSortedL(*saveSet));

  return(changes);
}

int eqList (struct list *l1, struct list *l2) {
  if (l1==l2) return(1);
  for(; l1 && l2; l1=l1->next, l2=l2->next) {
    if(l1->elem != l2->elem) return(0); /* diff elem */
  }
  if(l1==l2) return(1);  /* same length */
  return(0);
}

int memberUpdate(struct variable *lval, struct set *set, int update) {
  int memFlag;
  
  DESKIP(set);

  memFlag = memberEL(lval, &set->baseElements, update);
  if ((!memFlag) && update) set->size++;
#ifdef LOGSETS
  if(!memFlag) printf("s %d %d\n", update ? "M" : "m", lval-vars, recoverSetIndex(set));
#endif
  return(memFlag);
}

/* cache for insertSet: has "from" already been added to "to" */
#define CACHESIZE (1<<19)
static struct {
  struct set *from;
  struct set *to;
} cache[CACHESIZE];  /* fix: reduce size ??? */

static int hits=0;
static int misses=0;
static int clashes=0;

int insertSetCache(struct set *from, struct set *to, int update) {
  int i, hash;
  int f = ((int)from)>>4;
  int t = ((int)to)>>4;
  int startHash = (f*67 + t) & (CACHESIZE-1);
  
  hash = startHash;
  for(i=7; cache[hash].from && i; i--) {
    if((cache[hash].from==from) && (cache[hash].to==to)) {
      hits++;
      return(1);  /* already there */
    }
    hash = (hash + 17) & (CACHESIZE-1);
  }

  misses++;
  if(update) {
    if(!i) {  /* insert it at start */
      hash = startHash;
      clashes++;
    }
    
    cache[hash].from=from;
    cache[hash].to=to;
  }
  return(0);
}

int insertSetStd(struct set *from, struct set *to) {
  /* treat "from" as an atomic element, and add it to "to"; no cleaning */
  int change=0;

  DESKIP(from);
  DESKIP(to);

  if((from==to) || insertSetCache(from, to, 0)) return(0);

  if (!memberEL(from, &to->sets, 1)) {
    to->size++;
    change=1;
  }
  assert(to->size == (listSize(to->baseElements) + listSize(to->sets)));
  
  return(change);
}

int insertSetClean(struct set *from, struct set *to) {
  /* same functionality of insertSetStd, but deskips elements as it inserts */
  int change=0;

  DESKIP(from);
  DESKIP(to);

  if((from==to) || insertSetCache(from, to, 0)) return(0);

  /* self cleaning insertion */
  {
    struct list **sl = &to->sets;
    struct list *l;
    struct list *newSets = 0;
    
    assert(isSortedL(to->sets));

    for(; l=*sl; ) {
      struct set *s1 = l->elem;
      struct set *s2;

      if(s1 > from) {  /* gone past where element should be */
	(*sl)=cons(from, l);
	to->size++;
	change = 1;
	break;
      }

      if(s2=s1->skip) {  /* s1 needs to be de-skipped */
	DESKIP(s2); 
	  
	*sl = l->next; /* drop element */
	freeCons(l);
	to->size--;
	if(!insertSetCache(s2, to, 0)) { 
	  newSets = cons(s2, newSets); /* save new element */
	}
	if(s2 == from) {
	  break;  /* found it */
	}

      } else {  /* s1 has no skip ptr */
	if(s1 == from) {
	  break;  /* found it */
	}
	sl = &(l->next);
      }
    }

    /* if got to end of list, add element at end */
    if(!l) {
      (*sl)=cons(from, l);
      to->size++;
      change = 1;
    }      

    /* now add back new elements */
    for(; newSets; ) {
      struct list *tmp=newSets;
      insertSetStd(newSets->elem, to);
      newSets=newSets->next;
      freeCons(tmp);
    }
  }

  assert(isSortedL(to->sets));
  insertSetCache(from, to, 1); /* don't assert its there until it is! */
  return(change);
}

int insertSet(struct set *from, struct set *to) {
  int change;
  change = insertSetClean(from, to);
#ifdef LOGSETS
	if(change) printf("S %d %d\n", from-sets, to-sets);
#endif
  return(change);
}
    
/* Hash table for finding identical sets. */
#define SAMESETSHASH (1<<14)  /* 8096 */
struct set *sameSets[SAMESETSHASH];

/* Look for a set with the same baseElements/sets. */
/* Todo: possible efficiency hacks 
   1. flush this hash table at end of each iteration.
   2. in combination with 1, don't update table?
      (better chance of reusing?)
 */
struct set *hashSet(struct set *s, int update) {
  /* looks for an equivalent set;
     if it finds one, then it returns the set
     in update mode, it caches s if can't find equivalent set */

  int hash;

  if(!s->size) return(s);  /* don't try to hash empty or skip sets */
  assert(!s->skip);

  hash = s->size & (SAMESETSHASH-1);

  if(sameSets[hash]) {
    struct set *s1 = sameSets[hash];
    if(eqList(s1->baseElements, s->baseElements) && eqList(s1->sets, s->sets)) { 
      return(s1);   /* found equivalent set */
    } else if((s1->size & (SAMESETSHASH-1)) == hash) {
      return(s);  /* if curr set is fresh, leave it be */
    }
  }
  if(update) {
    sameSets[hash] = s;  /* save pointer to s */
  }
  return(s);
}

/***********************************************
* Management of lval lists.                   
* Main problems: 
*    - want to share  
*    - want to garbage collect
*
* Interface:
*  manageLvalList: hand-over management of a list.
*  gcLvalLists: assume no-one wants the lists anymore and throw them away.
*
* 1. hash table for finding identical lval lists 
* 2. a list to keep lists not in hash table. 
*************************************************/

#define SAMELVALSHASH (1<<14)  /* 8096 */
struct list *lvalLists[SAMELVALSHASH];

struct list *manageLvalList(struct list *l) {
  int hash = listSize(l);
  struct list *ll;
  int i;

  hash = hash & (SAMELVALSHASH-1);

  for(ll = lvalLists[hash], i=3; ll && i; ll=ll->next, i--) {
    struct list *l1 = ll->elem;
    if(eqList(l1, l)) {
      freeL(l);
      return(l1);   /* found equivalent set */
    }
  }

  {
    lvalLists[hash] = cons(l, lvalLists[hash]);
    return(l);
  }
}

void gcLvalLists () { /* reset manageLvalList table */
  int i;
  for(i=0; i<SAMELVALSHASH; i++) {
    struct list *l = lvalLists[i];
    lvalLists[i] = 0;
    for(; l; ) {
      struct list *l1 = l;
      freeL(l->elem);
      l=l->next;
      freeCons(l1);
    }
  }
}


int skips=0;

struct list *freedUpLists = 0;

void unifyNodes(struct set *s1, struct set *s2) { /* sets skip of deskip(s1) to deskip(s2) */
  DESKIP(s1);
  DESKIP(s2);

  if(s1 == s2) return;  /* already unified */

  assert(!s2->skip);
  s1->skip = s2;
  skips++;
  s2->size += subsetL(s1->baseElements, &s2->baseElements, 1)
            + subsetL(s1->sets, &s2->sets, 1);

  /* NB: cannot free up baseElements and sets just yet;
     calls to getLvals higher up on the stack still
     refer to these lists; delay freeing them
     until we finish the top level call to getLvals */
  freedUpLists = cons(s1->baseElements, freedUpLists);
  freedUpLists = cons(s1->sets, freedUpLists);
  /* safe to set the following to zero because 
     they will never be inspected: since skip field is set */
  s1->baseElements = 0;
  s1->sets = 0;
  s1->size = 0;
  if(s1->lvalLvals) {
    if(s2->lvalLvals) {
      unifyNodes(s1->lvalLvals, s2->lvalLvals);
    } else {
      s2->lvalLvals=s1->lvalLvals;
    }
  }
}
  
/** node status
  
  stale: haven't visited this node during current iteration
         (LAST_RECOMPUTED(s) != currentIterationIndex)
  onpath: we're still in the middle of working on this node
          LAST_RECOMPUTED(s) == currentIterationIndex && s->onPath
  done: finished visiting; info is all correct
          LAST_RECOMPUTED(s) == currentIterationIndex && !s->onPath
**/

struct list *getLvals(struct set *s, struct list *path) {

  if(!s) return(0);
  DESKIP(s);
  assert(!s->skip);

  if(LAST_RECOMPUTED(s) == currentIterationIndex) {
    assert(!s->skip);
    if(s->onPath) {  /* we're still visiting this node: found cycle! */
      struct list *l;
      assert(!s->skip);
      for(l=path; l && (l->elem != s); l=l->next) {
	struct set *s1 = l->elem;
	unifyNodes(s1, s);
      }
      assert(l);
      return(0);
    } else {  /* !(s->onPath): we've finished that computation */
      return(s->lvals);
    }

  } else { /* LAST_RECOMPUTED(s) != currentIterationIndex */
    struct set *s1 = hashSet(s, 0);  /* returns s or an equivalent set */
    if((s1 != s) && (LAST_RECOMPUTED(s1) == currentIterationIndex) && !s1->onPath) {
      /* we've finished recomputing s1, an equivalent set to s */
      return(s1->lvals);
    } else {  /* recompute s by visiting all of the children of s */
      struct list *lvals=0;
      struct list *l;
      struct list *new = cons(s, path);
      SET_LAST_RECOMPUTED(s, currentIterationIndex);
      s->lvals = 0;  /* throw away old lvals; gc handled by gcLvalLists */
      s->onPath=1;
      
      subsetL(s->baseElements, &lvals, 1);
      for(l=s->sets; l; l=l->next) {
	subsetL(getLvals((struct set*) l->elem, new), &lvals, 1);
      }
      freeCons(new);
      s->onPath=0;  /* also used to indicate we've finished with this node */
      hashSet(s, 1);  /* cache results in hashSet */
      lvals = manageLvalList(lvals);
      s->lvals = lvals;
      return(lvals);
    }
  }
}

void stats() {
  int i;
  int totSize = 0, baseTot = 0, setsTot = 0, lvalTot = 0, programVarLvals = 0;
  int depSets;
 
  printf("sets: %d\n", nextSet);

  for(i=0; i<nextSet; i++) {
    struct set *s = sets[chunkIndex(i)]+chunkOffset(i);
    int i1 = listSize(s->baseElements);
    int i2 = listSize(s->sets);
    int lvalsCount;
    
    totSize += i1 + i2;
    baseTot += i1;
    setsTot += i2;
    lvalsCount = listSize(getElems(s));
    lvalTot += lvalsCount;
  }

  depSets = 0; /* getDepSetsCount(); fix */

  printf("Sets: Base=%d Sets=%d+%d Tot=%d Lvals=%d\n", baseTot, depSets, setsTot-depSets, totSize, lvalTot);
  totSize = 0, baseTot = 0, setsTot = 0, lvalTot = 0;

  for(i=0; i<varCount(); i++) {
    struct variable *v = varOfIndex(i);
    struct set *s;
    if(s=v->lvals) {
      int i1 = listSize(s->baseElements);
      int i2 = listSize(s->sets);
      int lvalsCount;
      
      totSize += i1 + i2;
      baseTot += i1;
      setsTot += i2;
      lvalsCount = listSize(getElems(s));
      lvalTot += lvalsCount;
      programVarLvals += lvalsCount;
    }
  }
  depSets = 0; /* getDepSetsCount(); fix */
  
  printf("Vars: Base=%d Sets=%d+%d Tot=%d Lvals=%d\n", baseTot, depSets, setsTot-depSets, totSize, lvalTot);
  printf("Variable Lvals: %d\n", programVarLvals);
}

void dumpSet(struct set *s) {
  printf("%d: \n", recoverSetIndex(s));
  printf("  baseElements = "); prL(s->baseElements);
  printf("  sets = "); prSetL(s->sets);
  printf("  size = %d\n", s->size);
/*  printf("  getElems = "); prL(getElems(s)); */
  printf("  skip = %d\n", recoverSetIndex(s));
  printf("  lastRecomputed = %d\n", LAST_RECOMPUTED(s));
  printf("  onPath = %d\n", s->onPath);
}
  
int setsCount () {
  return(nextSet);
}

int setSizes() {
  if(debug) printf("(cache: %dh+%df+%dc/%d)\n", hits, misses-clashes, clashes, hits+misses);
  return(0);
}

int computeSetSizes() {

  int i;
  int totSize = 0;
  for(i=0; i<nextSet; i++) {
    struct set *s = sets[chunkIndex(i)] + chunkOffset(i);
    totSize += (listSize(getElems(s)));
  }

  if(debug) printf("[%d]", totSize);
  return(totSize);
}

void recomputeElemElems(struct list *lvals, struct set *lvalLvals) {
  for(; lvals; lvals=lvals->next) {
    struct variable *lval = lvals->elem;
    if(lval->lvals) {
      insertSet(lval->lvals, lvalLvals);
    }
  }
}

struct list *getElems(struct set *s) { /* wrapper function for getting lvals in set */
  struct list *lvals;
  if(!s) return(0);
  DESKIP(s);
  if(LAST_RECOMPUTED(s) == currentIterationIndex) {
    return(s->lvals);
  }

  freedUpLists = 0;
  lvals = getLvals(s, 0);
  freeLL(freedUpLists);
  assert(isSortedL(lvals));
  
  return(lvals);
}

struct set *getElemElems(struct set *s) { /* get set representing lvals in each elt. of lvals->s */
  struct list *lvals;
  
  if(!s) return(0);
  DESKIP(s); 

  if(!s->lvalLvals) {
    s->lvalLvals = emptySet();
  }

  if(LAST_RECOMPUTED(s) == currentIterationIndex) {
    if(LVAL_LVALS_RECOMPUTED(s)) {
      return(s->lvalLvals);
    }
    lvals=s->lvals;
  } else {
    lvals = getElems(s);
  }

  recomputeElemElems(lvals, s->lvalLvals);
  SET_LVAL_LVALS_RECOMPUTED(s);
  return(s->lvalLvals);  /* NB: may not be deskipped */
}
