(**********************************************************
 * This software is provided subject to the following terms
 * and conditions, which you should read carefully before
 * using the software. Using this software indicates your
 * acceptance of these terms and conditions.  If you do not
 * agree with these terms and conditions, do not use the
 * software. 
 * 
 * Copyright (c) 2002, Agere Systems
 * All rights reserved. 
 * 
 * Redistribution and use in source or binary forms, with or
 * without modifications, are permitted provided that the
 * following conditions are met: 
 * 
 * o Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following Disclaimer in comments in the code as well as
 *   in the documentation and/or other materials provided with
 *   the distribution.
 * o Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   Disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * o Neither the name of Agere Systems nor the names of the
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission. 
 * 
 * Disclaimer 
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, INFRINGEMENT AND THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. ANY USE, MODIFICATION OR DISTRIBUTION OF
 * THIS SOFTWARE IS SOLELY AT THE USER'S OWN RISK. IN NO EVENT SHALL
 * AGERE SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, INCLUDING, BUT NOT LIMITED
 * TO, CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 **********************************************************)

structure DriveTrack =
  struct
    val skippedFiles = ref 0
      
    val mallocs = ref(nil:string list)
    val noFields = ref false
    val noStrings = ref false
    val commonTypeFields = ref false
    val commonSizeFields = ref false
    val alwaysTrackLvals = ref false
    val outputDir = ref (NONE : string option)
    val debug = ref false
    val pids = ref false
      
    fun parseFile file =
      (print ("\nParsing file " ^ file ^ ".\n");
       let val {ast, tidtab, errorCount, auxiliaryInfo={aidtab, ...}, ...} = ParseToAst.fileToAst file
	 val outputFile =
	   if (String.substring(file, (String.size file)-2, 2)) = ".c"
	     orelse (String.substring(file, (String.size file)-2, 2)) = ".i"
	     then (String.substring(file, 0, (String.size file)-2)) ^ ".gda"
	   else file ^ ".gda"
	 val outputFile = (case !outputDir of
			     NONE => outputFile
			   | SOME dir => let val {file, ...} = OS.Path.splitDirFile outputFile
					 in
					   OS.Path.joinDirFile {dir=dir, file=file} 
					 end)
	 val outStream = TextIO.openOut outputFile
	 val flags = {mallocs=(!mallocs), noFields=(!noFields),
		      commonSizeFields=(!commonSizeFields),
		      commonTypeFields=(!commonTypeFields),
		      alwaysTrackLvals=(!alwaysTrackLvals),
		      noStrings=(!noStrings)}
	 val (vars, statics, consts) = Track.travAst (tidtab, aidtab) flags ast (* parse the AST *)
	 val _ = GDAFile.writeGDAFile outStream tidtab (!debug) (!pids) statics vars consts
	 (* dump dependencies *)
	 val _ = TextIO.closeOut outStream
	 val _ = (case errorCount of
		    0 => ()
		  | i => print ("gdac completed with " ^ (Int.toString i) ^ " errors\n"))

       in
	 ()
       end
     handle xxx => (print ("Parse problem: skipping " ^ file ^ "\n");
		    print "Exception: ";
		    List.map (fn x => (print x; print ".")) (SMLofNJ.exnHistory xxx);
		    print "\n";
		    skippedFiles := !skippedFiles + 1))
      
    fun go files =
      ( (BuildAst.local_externs_ok := true;
	 skippedFiles := 0;
	 List.map parseFile files;
	 if (!skippedFiles > 0) then
	   print("Skipped " ^ (Int.toString (!skippedFiles)) ^ " files because of parse problems\n")
	 else ();
	   OS.Process.success)
       handle xxx => (print "Exception: ";
		      List.map (fn x => (print x;
					 print ".")) (SMLofNJ.exnHistory xxx);
		      print "\n";
		      OS.Process.failure) )

    fun main(_, l) =
      let 
	fun errorMsg () = (print "Usage: gdac -f configfile or gdac file\n";
			   print "Use -help flag for more options\n";
			   OS.Process.failure)
	val l = (case l of  (* stip off command name *)
		   _ :: l => l
		 | nil => nil)
	fun readArgs ("-malloc" :: mallocName :: l) = (mallocs := mallocName :: (!mallocs);
						       readArgs l)
	  | readArgs ("-noFields" :: l) = (noFields := true;
					   readArgs l)
	  | readArgs ("-commonTypeFields" :: l) = (commonTypeFields := true;
						   readArgs l)
	  | readArgs ("-commonSizeFields" :: l) = (commonSizeFields := true;
						   readArgs l)
	  | readArgs ("-alwaysTrackLvals" :: l) = (alwaysTrackLvals := true;
						   readArgs l)
	  | readArgs ("-debug" :: l) = (debug := true;
					readArgs l)
	  | readArgs ("-pid" :: l) = (pids := true;
					readArgs l)
	  | readArgs ("-noStrings" :: l) = (noStrings := true;
					    readArgs l)
	  | readArgs ("-outputDir" :: dir :: l) = (outputDir := SOME dir;
						   if (OS.FileSys.isDir dir handle _ => false) then readArgs l
						   else (print "Bad outputDir: directory doesn't exist.\n";
							 readArgs ["-help"]))
	  | readArgs ("-help" :: _) =
	  (print "Usage: gdac -f configfile or gdac file\n";
	   print "Additional Flags:\n";
	   print "   -malloc mallocName: treats function mallocName as a malloc function\n";
	   print "                       each location of a call to mallocName generates a new lval.\n";
	   print "\n";
	   print "   -noFields:          field of structs and unions are ignored\n";
	   print "                       e.g. e.f1, e.f2, ... are treated as e.\n";
	   print "\n";
	   print "   -commonTypeFields:  common initial sequences of fields in structs are identified\n";
	   print "                       (common initial sequence == same sequence of types.\n";
	   print "\n";
	   print "   -commonSizeFields:  common initial sequences of fields in structs are identified\n";
	   print "                       (common initial sequence == same sequence of types sizes.\n";
	   print "\n";
	   print "   -alwaysTrackLvals:  track lvals through all arithmetic operations.\n";
	   print "\n";
	   print "   -outputDir dir:     creates .gda files in directory dir.\n";
	   print "\n";
	   print "   -debug:             dumps dependencies on stdout.\n";
	   print "\n";
	   print "   -pid:               uses pids instead of names.\n";
	   print "\n";
	   print "   -noStrings:         ignores strings.\n";
	   OS.Process.success)
	  | readArgs ["-f", configfile] = (go (readConfig.readConfig configfile))
	  | readArgs [file] = go [file]
	  | readArgs _ = errorMsg()
      in
	readArgs l
      end 

  end;

val _ = SMLofNJ.exportFn ("gdac", DriveTrack.main);
